"""
This script randomly generates earth-style models, then randomly generates data
from those models and fits earth models to those data using both the python and
R implementations.  It records the sample size, m, the number of input
dimensions, n, the number of forward pass iterations, the runtime, and the
r^2 statistic for each fit and writes the result to a CSV file.
"""

from __future__ import annotations

import time

import numpy
import pandas
import pandas.rpy.common as com
import rpy2.robjects as robjects

from pyearth import Earth


class DataGenerator(object):
    def __init__(self):
        pass

    def generate(self, m):
        pass


class NoiseGenerator(DataGenerator):
    def __init__(self, n):
        self.n = n

    def generate(self, m):
        X = numpy.random.normal(size=(m, self.n))
        y = numpy.random.normal(size=m)
        return X, y


class LinearGenerator(DataGenerator):
    def __init__(self, n):
        self.n = n

    def generate(self, m):
        X = numpy.random.normal(size=(m, self.n))
        beta = numpy.random.normal(size=self.n)
        y = numpy.dot(X, beta) + numpy.random.normal(m)
        return X, y


class VFunctionGenerator(DataGenerator):
    def __init__(self, n):
        self.n = n

    def generate(self, m):
        X = numpy.random.normal(size=(m, self.n))
        var = numpy.random.randint(self.n)
        y = 10 * abs(X[:, var]) + numpy.random.normal(m)
        return X, y


class UFunctionGenerator(DataGenerator):
    def __init__(self, n):
        self.n = n

    def generate(self, m):
        X = numpy.random.normal(size=(m, self.n))
        var = numpy.random.randint(self.n)
        y = 10 * (X[:, var] ** 2) + numpy.random.normal(m)
        return X, y


class RandomComplexityGenerator(DataGenerator):
    def __init__(self, n, max_terms=10, max_degree=2):
        self.n = n
        self.max_terms = max_terms
        self.max_degree = max_degree

    def generate(self, m):
        X = numpy.random.normal(size=(m, self.n))
        # Including the intercept
        num_terms = numpy.random.randint(2, self.max_terms)
        coef = 10 * numpy.random.normal(size=num_terms)
        B = numpy.ones(shape=(m, num_terms))
        B[:, 0] += coef[0]
        for i in range(1, num_terms):
            degree = numpy.random.randint(1, self.max_degree)
            for _bf in range(degree):
                knot = numpy.random.normal()
                dir = 1 - 2 * numpy.random.binomial(1, 0.5)
                var = numpy.random.randint(0, self.n)
                B[:, i] *= (dir * (X[:, var] - knot)) * (dir * (X[:, var] - knot) > 0)
        y = numpy.dot(B, coef) + numpy.random.normal(size=m)
        return X, y


def run_earth(X, y, **kwargs):
    """
    Run with the R package earth.
    Return prediction value, training time, and number
    of forward pass iterations.
    """
    r = robjects.r
    m, n = X.shape
    data = pandas.DataFrame(X)
    data["y"] = y
    r_data = com.convert_to_r_dataframe(data)
    r("library(earth)")
    r_func = """
    run <-  function(data, degree=1, fast.k=0, penalty=3.0){
       time = system.time(model <- earth(y~.,data=data,degree=degree,penalty=penalty))[3]
       forward_terms = dim(summary(model)$prune.terms)[1]
       y_pred = predict(model,data)
       return(list(y_pred, time, forward_terms, model))
     }
     """
    r(r_func)
    run = r("run")
    r_list = run(
        **{
            "data": r_data,
            "degree": kwargs["max_degree"],
            "fast.k": 0,
            "penalty": kwargs["penalty"],
        }
    )
    y_pred = numpy.array(r_list[0]).reshape(m)
    time = r_list[1][0]
    forward_terms = r_list[2][0]
    return y_pred, time, (forward_terms - 1) / 2


def run_pyearth(X, y, **kwargs):
    """
    Run with pyearth.
    Return prediction value, training time, and number of
    forward pass iterations.
    """
    model = Earth(**kwargs)
    t0 = time.time()
    model.fit(X, y)
    t1 = time.time()
    y_pred = model.predict(X)
    forward_iterations = len(model.forward_trace()) - 1
    return y_pred, t1 - t0, forward_iterations


def compare(generator_class, sample_sizes, dimensions, repetitions, **kwargs):
    """
    Return a data table that includes m, n, pyearth or earth, training time,
    and number of forward pass iterations.
    """
    header = ["m", "n", "pyearth", "earth", "time", "forward_iterations", "rsq"]
    data = []
    for n in dimensions:
        generator = generator_class(n=n)
        for m in sample_sizes:
            for rep in range(repetitions):
                print(n, m, rep)
                X, y = generator.generate(m=m)
                y_pred_r, time_r, iter_r = run_earth(X, y, **kwargs)
                rsq_r = 1 - (numpy.sum((y - y_pred_r) ** 2)) / (numpy.sum((y - numpy.mean(y)) ** 2))
                data.append([m, n, 0, 1, time_r, iter_r, rsq_r])
                y_pred_py, time_py, iter_py = run_pyearth(X, y, **kwargs)
                rsq_py = 1 - (numpy.sum((y - y_pred_py) ** 2)) / (numpy.sum((y - numpy.mean(y)) ** 2))
                data.append([m, n, 1, 0, time_py, iter_py, rsq_py])
    return pandas.DataFrame(data, columns=header)


if __name__ == "__main__":
    sample_sizes = [100, 200, 300, 500]
    dimensions = [10, 20, 30]
    rep = 5
    numpy.random.seed(1)
    data = compare(
        RandomComplexityGenerator,
        sample_sizes,
        dimensions,
        rep,
        max_degree=2,
        penalty=3.0,
    )
    print(data)
    data.to_csv("comparison.csv")



"""
====================================
Plotting the absolute value function
====================================

A simple example plotting a fit of the absolute value function.

"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy

from pyearth import Earth

# Create some fake data
numpy.random.seed(2)
m = 1000
n = 10
X = 80 * numpy.random.uniform(size=(m, n)) - 40
y = numpy.abs(X[:, 6] - 4.0) + 1 * numpy.random.normal(size=m)

# Fit an Earth model
model = Earth(max_degree=1, verbose=True)
model.fit(X, y)

# Print the model
print(model.trace())
print(model.summary())

# Plot the model
y_hat = model.predict(X)
plt.figure()
plt.plot(X[:, 6], y, "r.")
plt.plot(X[:, 6], y_hat, "b.")
plt.show()



# distutils: language = c
# cython: cdivision = True
# cython: boundscheck = False
# cython: wraparound = False
# cython: profile = False

from ._record cimport PruningPassIteration
from ._util cimport gcv, apply_weights_2d
import numpy as np
from scipy.linalg import lstsq
from collections import defaultdict

GCV, RSS, NB_SUBSETS = "gcv", "rss", "nb_subsets"
FEAT_IMP_CRITERIA = (GCV, RSS, NB_SUBSETS)

cdef class PruningPasser:
    '''Implements the generic pruning pass as described by Friedman, 1991.'''
    def __init__(PruningPasser self, Basis basis,
                 cnp.ndarray[FLOAT_t, ndim=2] X,
                 cnp.ndarray[BOOL_t, ndim=2] missing,
                 cnp.ndarray[FLOAT_t, ndim=2] y,
                 cnp.ndarray[FLOAT_t, ndim=2] sample_weight, int verbose,
                 **kwargs):
        self.X = X
        self.missing = missing
        self.m = self.X.shape[0]
        self.n = self.X.shape[1]
        self.y = y
        self.sample_weight = sample_weight
        self.verbose = verbose
        self.basis = basis
        self.B = np.empty(shape=(self.m, len(self.basis) + 1), dtype=np.float64)
        self.penalty = kwargs.get('penalty', 3.0)
        if sample_weight.shape[1] == 1:
            y_avg = np.average(self.y, weights=sample_weight[:,0], axis=0)
        else:
            y_avg = np.average(self.y, weights=sample_weight, axis=0)

        # feature importance
        feature_importance_criteria = kwargs.get("feature_importance_type", [])
        if isinstance(feature_importance_criteria, basestring):
            feature_importance_criteria = [feature_importance_criteria]
        self.feature_importance = dict()
        for criterion in feature_importance_criteria:
            self.feature_importance[criterion] = np.zeros((self.n,))

    cpdef run(PruningPasser self):
        # This is a totally naive implementation and could potentially be made
        # faster through the use of updating algorithms.  It is not clear that
        # such optimization would be worthwhile, as the pruning pass is not the
        # slowest part of the algorithm.
        cdef INDEX_t i
        cdef INDEX_t j
        cdef long v
        cdef INDEX_t basis_size = len(self.basis)
        cdef INDEX_t pruned_basis_size = self.basis.plen()
        cdef FLOAT_t gcv_
        cdef INDEX_t best_iteration
        cdef INDEX_t best_bf_to_prune
        cdef FLOAT_t best_gcv
        cdef FLOAT_t best_iteration_gcv
        cdef FLOAT_t best_iteration_mse
        cdef FLOAT_t mse, mse0, total_weight

        cdef cnp.ndarray[FLOAT_t, ndim = 2] B = (
            <cnp.ndarray[FLOAT_t, ndim = 2] > self.B)
        cdef cnp.ndarray[FLOAT_t, ndim = 2] X = (
            <cnp.ndarray[FLOAT_t, ndim = 2] > self.X)
        cdef cnp.ndarray[BOOL_t, ndim = 2] missing = (
            <cnp.ndarray[BOOL_t, ndim = 2] > self.missing)
        cdef cnp.ndarray[FLOAT_t, ndim = 2] y = (
            <cnp.ndarray[FLOAT_t, ndim = 2] > self.y)
        cdef cnp.ndarray[FLOAT_t, ndim = 2] sample_weight = (
            <cnp.ndarray[FLOAT_t, ndim = 2] > self.sample_weight)
        cdef cnp.ndarray[FLOAT_t, ndim = 1] weighted_y

        if self.verbose >= 1:
            print('Beginning pruning pass')

        # Initial solution
        mse = 0.
        mse0 = 0.
        total_weight = 0.
        for p in range(y.shape[1]):
            if sample_weight.shape[1] == 1:
                weighted_y = y[:,p] * np.sqrt(sample_weight[:,0])
                self.basis.weighted_transform(X, missing, B, sample_weight[:, 0])
                total_weight += np.sum(sample_weight[:,0])
                mse0 += np.sum(sample_weight[:,0] * (y[:,p] - np.average(y[:,p], weights=sample_weight[:,0])) ** 2)
            else:
                weighted_y = y[:,p] * np.sqrt(sample_weight[:,p])
                self.basis.weighted_transform(X, missing, B, sample_weight[:, p])
                total_weight += np.sum(sample_weight[:,p])
                mse0 += np.sum(sample_weight[:,p] * (y[:,p] - np.average(y[:,p], weights=sample_weight[:,p])) ** 2)
            if sample_weight.shape[1] == 1:
                self.basis.weighted_transform(X, missing, B, sample_weight[:, 0])
            else:
                self.basis.weighted_transform(X, missing, B, sample_weight[:, p])
            beta, mse_ = lstsq(B[:, 0:(basis_size)], weighted_y)[0:2]
            if mse_.size:
                pass
            else:
                mse_ = np.sum(
                    (np.dot(B[:, 0:basis_size], beta) - weighted_y) ** 2)
            mse += mse_

        # Create the record object
        self.record = PruningPassRecord(
            self.m, self.n, self.penalty, mse0 / total_weight, pruned_basis_size, mse / total_weight)
        gcv_ = self.record.gcv(0)
        best_gcv = gcv_
        best_iteration = 0

        if self.verbose >= 1:
            print(self.record.partial_str(slice(-1, None, None), print_footer=False))

        # init feature importance
        prev_best_iteration_gcv = None
        prev_best_iteration_mse = None

        # Prune basis functions sequentially
        for i in range(1, pruned_basis_size):
            first = True
            pruned_basis_size -= 1

            # Find the best basis function to prune
            for j in range(basis_size):
                bf = self.basis[j]
                if bf.is_pruned():
                    continue
                if not bf.is_prunable():
                    continue
                bf.prune()


                mse = 0.
                for p in range(y.shape[1]):
                    if sample_weight.shape[1] == 1:
                        weighted_y = y[:,p] * np.sqrt(sample_weight[:,0])
                        self.basis.weighted_transform(X, missing, B, sample_weight[:, 0])
                    else:
                        weighted_y = y[:,p] * np.sqrt(sample_weight[:,p])
                        self.basis.weighted_transform(X, missing, B, sample_weight[:, p])
                    beta, mse_ = lstsq(
                        B[:, 0:pruned_basis_size], weighted_y)[0:2]
                    if mse_.size:
                        pass
#                         mse_ /= np.sum(self.sample_weight)
                    else:
                        mse_ = np.sum((np.dot(B[:, 0:pruned_basis_size], beta) -
                                    weighted_y) ** 2) #/ np.sum(sample_weight)
                    mse += mse_# * output_weight[p]
                gcv_ = gcv(mse / np.sum(sample_weight), pruned_basis_size, self.m, self.penalty)

                if gcv_ <= best_iteration_gcv or first:
                    best_iteration_gcv = gcv_
                    best_iteration_mse = mse
                    best_bf_to_prune = j
                    first = False
                bf.unprune()

            # Feature importance
            if i > 1:
                # having selected the best basis to prune, we compute how much
                # that basis decreased the mse and gcv relative to the previous mse and gcv
                # respectively.
                mse_decrease = (best_iteration_mse - prev_best_iteration_mse)
                gcv_decrease = (best_iteration_gcv - prev_best_iteration_gcv)
                variables = set()
                bf = self.basis[best_bf_to_prune]
                for v in bf.variables():
                    variables.add(v)
                for v in variables:
                    if RSS in self.feature_importance:
                        self.feature_importance[RSS][v] += mse_decrease
                    if GCV in self.feature_importance:
                        self.feature_importance[GCV][v] += gcv_decrease
                    if NB_SUBSETS in self.feature_importance:
                        self.feature_importance[NB_SUBSETS][v] += 1
            # The inner loop found the best basis function to remove for this
            # iteration. Now check whether this iteration is better than all
            # the previous ones.
            if best_iteration_gcv <= best_gcv:
                best_gcv = best_iteration_gcv
                best_iteration = i

            prev_best_iteration_gcv = best_iteration_gcv
            prev_best_iteration_mse = best_iteration_mse
            # Update the record and prune the selected basis function
            self.record.append(PruningPassIteration(
                best_bf_to_prune, pruned_basis_size, best_iteration_mse / total_weight))
            self.basis[best_bf_to_prune].prune()

            if self.verbose >= 1:
                print(self.record.partial_str(slice(-1, None, None), print_header=False, print_footer=(pruned_basis_size == 1)))

        # Unprune the basis functions pruned after the best iteration
        self.record.set_selected(best_iteration)
        self.record.roll_back(self.basis)
        if self.verbose >= 1:
            print(self.record.final_str())

        # normalize feature importance values
        for name, val in self.feature_importance.items():
            if name == 'gcv':
                val[val < 0] = 0 # gcv can have negative feature importance corresponding to an increase of gcv, set them to zero
            if val.sum() > 0:
                val /= val.sum()
            self.feature_importance[name] = val

    cpdef PruningPassRecord trace(PruningPasser self):
        return self.record



# distutils: language = c
# cython: cdivision = True
# cython: boundscheck = False
# cython: wraparound = False
# cython: profile = False

from ._util cimport gcv_adjust, log2, apply_weights_1d, apply_weights_slice
from ._basis cimport (Basis, BasisFunction, ConstantBasisFunction,
                      HingeBasisFunction, LinearBasisFunction,
                      MissingnessBasisFunction)
from ._record cimport ForwardPassIteration
from ._types import BOOL, INT
from ._knot_search cimport knot_search, MultipleOutcomeDependentData, PredictorDependentData, \
    KnotSearchReadOnlyData, KnotSearchState, KnotSearchWorkingData, KnotSearchData
import sys
from libc.math cimport sqrt, abs, log
import numpy as np
cnp.import_array()

from heapq import heappush, heappop
class FastHeapContent:

    def __init__(self, idx, mse=-np.inf, m=-np.inf, v=None):
        """
        This class defines an entry of the priority queue as defined in [1].
        The entry stores information about parent basis functions and is
        used by the priority queue in the forward pass
        to choose the next parent basis function to try.

        References
        ----------
        .. [1] Fast MARS, Jerome H.Friedman, Technical Report No.110, May 1993.

        """
        self.idx = idx
        self.mse = mse
        self.m = m
        self.v = v

    def __lt__(self, other):
        return self.mse < other.mse

cdef int MAXTERMS = 0
cdef int MAXRSQ = 1
cdef int NOIMPRV = 2
cdef int LOWGRSQ = 3
cdef int NOCAND = 4
stopping_conditions = {
    MAXTERMS: "Reached maximum number of terms",
    MAXRSQ: "Achieved RSQ value within threshold of 1",
    NOIMPRV: "Improvement below threshold",
    LOWGRSQ: "GRSQ too low",
    NOCAND: "No remaining candidate knot locations"
}

cdef class ForwardPasser:

    def __init__(ForwardPasser self, cnp.ndarray[FLOAT_t, ndim=2] X,
                 cnp.ndarray[BOOL_t, ndim=2] missing,
                 cnp.ndarray[FLOAT_t, ndim=2] y,
                 cnp.ndarray[FLOAT_t, ndim=2] sample_weight,
                 **kwargs):

        cdef INDEX_t i
        self.X = X
        self.missing = missing
        self.y = y
        # Assuming Earth.fit got capital W (the inverse of squared variance)
        # so the objective function is (sqrt(W) * residual) ^ 2)
        self.sample_weight = np.sqrt(sample_weight)
        self.m = self.X.shape[0]
        self.n = self.X.shape[1]
        self.endspan       = kwargs.get('endspan', -1)
        self.minspan       = kwargs.get('minspan', -1)
        self.endspan_alpha = kwargs.get('endspan_alpha', .05)
        self.minspan_alpha = kwargs.get('minspan_alpha', .05)
        self.max_terms     = kwargs.get('max_terms', min(2 * self.n + self.m // 10, 400))
        self.allow_linear  = kwargs.get('allow_linear', True)
        self.max_degree    = kwargs.get('max_degree', 1)
        self.thresh        = kwargs.get('thresh', 0.001)
        self.penalty       = kwargs.get('penalty', 3.0)
        self.check_every   = kwargs.get('check_every', -1)
        self.min_search_points = kwargs.get('min_search_points', 100)
        self.xlabels       = kwargs.get('xlabels')
        self.use_fast = kwargs.get('use_fast', False)
        self.fast_K = kwargs.get("fast_K", 5)
        self.fast_h = kwargs.get("fast_h", 1)
        self.zero_tol = kwargs.get('zero_tol', 1e-12)
        self.allow_missing = kwargs.get("allow_missing", False)
        self.verbose = kwargs.get("verbose", 0)
        if self.allow_missing:
            self.has_missing = np.any(self.missing, axis=0).astype(BOOL)

        self.fast_heap = []

        if self.xlabels is None:
            self.xlabels = ['x' + str(i) for i in range(self.n)]
        if self.check_every < 0:
            self.check_every = (<int > (self.m / self.min_search_points)
                                if self.m > self.min_search_points
                                else 1)

        weighted_mean = np.mean((self.sample_weight ** 2) * self.y)
        self.sst = np.sum((self.sample_weight * (self.y - weighted_mean)) ** 2)
        self.basis = Basis(self.n)
        self.basis.append(ConstantBasisFunction())
        if self.use_fast is True:
            content = FastHeapContent(idx=0)
            heappush(self.fast_heap, content)

        self.mwork = np.empty(shape=self.m, dtype=np.int64)

        self.B = np.ones(
            shape=(self.m, self.max_terms + 4), order='F', dtype=np.float64)
        self.basis.transform(self.X, self.missing, self.B[:,0:1])

        if self.endspan < 0:
            self.endspan = round(3 - log2(self.endspan_alpha / self.n))

        self.linear_variables = np.zeros(shape=self.n, dtype=INT)
        self.init_linear_variables()

        # Removed in favor of new knot search code
        self.iteration_number = 0

        # Add in user selected linear variables
        for linvar in kwargs.get('linvars',[]):
            if linvar in self.xlabels:
                self.linear_variables[self.xlabels.index(linvar)] = 1
            elif linvar in range(self.n):
                self.linear_variables[linvar] = 1
            else:
                raise IndexError(
                    'Unknown variable selected in linvars argument.')

        # Initialize the data structures for knot search
        self.n_outcomes = self.y.shape[1]
        n_predictors = self.X.shape[1]
        n_weights = self.sample_weight.shape[1]
        self.workings = []
        self.outcome = MultipleOutcomeDependentData.alloc(self.y, self.sample_weight, self.m,
                                                          self.n_outcomes, self.max_terms + 4,
                                                          self.zero_tol)
        self.outcome.update_from_array(self.B[:,0])
        self.total_weight = 0.
        for i in range(self.n_outcomes):
            working = KnotSearchWorkingData.alloc(self.max_terms + 4)
            self.workings.append(working)
            self.total_weight += self.outcome.outcomes[i].weight.total_weight
        self.predictors = []
        for i in range(n_predictors):
            x = self.X[:, i].copy()
            x[missing[:,i]==1] = 0.
            predictor = PredictorDependentData.alloc(x)
            self.predictors.append(predictor)

        # Initialize the forward pass record
        self.record = ForwardPassRecord(
            self.m, self.n, self.penalty, self.outcome.mse(), self.xlabels)

    cpdef Basis get_basis(ForwardPasser self):
        return self.basis

    cpdef init_linear_variables(ForwardPasser self):
        cdef INDEX_t variable
        cdef cnp.ndarray[INT_t, ndim = 1] order
        cdef cnp.ndarray[INT_t, ndim = 1] linear_variables = (
            <cnp.ndarray[INT_t, ndim = 1] > self.linear_variables)
        cdef cnp.ndarray[FLOAT_t, ndim = 2] B = (
            <cnp.ndarray[FLOAT_t, ndim = 2] > self.B)
        cdef cnp.ndarray[FLOAT_t, ndim = 2] X = (
            <cnp.ndarray[FLOAT_t, ndim = 2] > self.X)
        cdef ConstantBasisFunction root_basis_function = self.basis[0]
        for variable in range(self.n):
            order = np.argsort(X[:, variable])[::-1].astype(np.int64)
            if root_basis_function.valid_knots(B[order, 0], X[order, variable],
                                               variable, self.check_every,
                                               self.endspan, self.minspan,
                                               self.minspan_alpha, self.n,
                                               self.mwork).shape[0] == 0:
                linear_variables[variable] = 1
            else:
                linear_variables[variable] = 0

    cpdef run(ForwardPasser self):
        if self.verbose >= 1:
            print('Beginning forward pass')
            print(self.record.partial_str(slice(-1, None, None), print_footer=False))
        if self.max_terms > 1 and self.record.mse(0) != 0.:
            while True:
                self.next_pair()
                if self.stop_check():
                    if self.verbose >= 1:
                        print(self.record.partial_str(slice(-1, None, None), print_header=False, print_footer=True))
                        print(self.record.final_str())
                    break
                else:
                    if self.verbose >= 1:
                        print(self.record.partial_str(slice(-1, None, None), print_header=False, print_footer=False))
                self.iteration_number += 1

    cdef stop_check(ForwardPasser self):
        last = self.record.__len__() - 1
        if self.record.iterations[last].get_size() > self.max_terms:
            self.record.stopping_condition = MAXTERMS
            return True
        rsq = self.record.rsq(last)
        if rsq > 1 - self.thresh:
            self.record.stopping_condition = MAXRSQ
            return True
        if last > 0:
            previous_rsq = self.record.rsq(last - 1)
            if rsq - previous_rsq < self.thresh:
                self.record.stopping_condition = NOIMPRV
                return True
        if self.record.grsq(last) < -10:
            self.record.stopping_condition = LOWGRSQ
            return True
        if self.record.iterations[last].no_further_candidates():
            self.record.stopping_condition = NOCAND
            return True
        if self.record.mse(last) == self.zero_tol:
            self.record.stopping_condition = NOIMPRV
            return True
        return False


    cpdef orthonormal_update(ForwardPasser self, b):
        # Update the outcome data
        linear_dependence = False
        return_codes = []
        return_code = self.outcome.update_from_array(b)
        if return_code == -1:
            raise ValueError('This should not have happened.')
        if return_code == 1:
            linear_dependence = True
        return linear_dependence

    cpdef orthonormal_downdate(ForwardPasser self):
        self.outcome.downdate()

    def trace(self):
        return self.record

    cdef next_pair(ForwardPasser self):
        cdef INDEX_t variable
        cdef INDEX_t parent_idx
        cdef INDEX_t parent_degree
        cdef INDEX_t nonzero_count
        cdef BasisFunction parent
        cdef cnp.ndarray[INDEX_t, ndim = 1] candidates_idx
        cdef FLOAT_t knot
        cdef FLOAT_t mse
        cdef INDEX_t knot_idx
        cdef FLOAT_t knot_choice
        cdef FLOAT_t mse_choice
        cdef FLOAT_t mse_choice_cur_parent
        cdef int variable_choice_cur_parent
        cdef int knot_idx_choice
        cdef INDEX_t parent_idx_choice
        cdef BasisFunction parent_choice
        cdef BasisFunction new_parent
        cdef BasisFunction new_basis_function
        parent_basis_content_choice = None
        parent_basis_content = None
        cdef INDEX_t variable_choice
        cdef bint first = True
        cdef bint already_covered
        cdef INDEX_t k = len(self.basis)
        cdef INDEX_t endspan
        cdef bint linear_dependence
        cdef bint dependent
        # TODO: Shouldn't there be weights here?
        cdef FLOAT_t gcv_factor_k_plus_1 = gcv_adjust(k + 1, self.m,
                                                      self.penalty)
        cdef FLOAT_t gcv_factor_k_plus_2 = gcv_adjust(k + 2, self.m,
                                                      self.penalty)
        cdef FLOAT_t gcv_factor_k_plus_3 = gcv_adjust(k + 3, self.m,
                                                      self.penalty)
        cdef FLOAT_t gcv_factor_k_plus_4 = gcv_adjust(k + 4, self.m,
                                                      self.penalty)
        cdef FLOAT_t gcv_
        cdef FLOAT_t mse_
        cdef INDEX_t i
        cdef bint eligible
        cdef bint covered
        cdef bint missing_flag
        cdef bint choice_needs_coverage

        cdef cnp.ndarray[FLOAT_t, ndim = 2] X = (
            <cnp.ndarray[FLOAT_t, ndim = 2] > self.X)
        cdef cnp.ndarray[BOOL_t, ndim = 2] missing = (
            <cnp.ndarray[BOOL_t, ndim = 2] > self.missing)
        cdef cnp.ndarray[FLOAT_t, ndim = 2] B = (
            <cnp.ndarray[FLOAT_t, ndim = 2] > self.B)
        cdef cnp.ndarray[INT_t, ndim = 1] linear_variables = (
            <cnp.ndarray[INT_t, ndim = 1] > self.linear_variables)
        cdef cnp.ndarray[BOOL_t, ndim = 1] has_missing = (
            <cnp.ndarray[BOOL_t, ndim = 1] > self.has_missing)
        cdef cnp.ndarray[FLOAT_t, ndim = 1] b
        cdef cnp.ndarray[FLOAT_t, ndim = 1] p
        cdef bint variable_can_be_linear

        if self.use_fast:
            nb_basis = min(self.fast_K, k, len(self.fast_heap))
        else:
            nb_basis = k

        content_to_be_repushed = []
        for idx in range(nb_basis):
            # Iterate over parents
            if self.use_fast:
                # retrieve the next basis function to try as parent
                parent_basis_content = heappop(self.fast_heap)
                content_to_be_repushed.append(parent_basis_content)
                parent_idx = parent_basis_content.idx
                mse_choice_cur_parent = -1
                variable_choice_cur_parent = -1
            else:
                parent_idx = idx

            parent = self.basis.get(parent_idx)
            if not parent.is_splittable():
                continue

            if self.use_fast:
                # each "fast_h" iteration, force to pass through all the variables,
                if self.iteration_number - parent_basis_content.m >= self.fast_h:
                    variables = range(self.n)
                    parent_basis_content.m = self.iteration_number
                # in the opposite case, just use the last chosen variable
                else:
                    if parent_basis_content.v is not None:
                        variables = [parent_basis_content.v]
                    else:
                        variables = range(self.n)
            else:
                variables = range(self.n)

            parent_degree = parent.effective_degree()

            for variable in variables:
                # Determine whether this variable can be linear
                variable_can_be_linear = self.allow_linear and not parent.has_linear(variable)

                # Determine whether missingness needs to be accounted for.
                if self.allow_missing and has_missing[variable]:
                    missing_flag = True
                    eligible = parent.eligible(variable)
                    covered = parent.covered(variable)
                else:
                    missing_flag = False

                # Make sure not to exceed max_degree (but don't count the
                # covering missingness basis function if required)
                if self.max_degree >= 0:
                    if parent_degree >= self.max_degree:
                        continue

                # If there is missing data and this parent is not
                # an eligible parent for this variable with missingness
                # (because it includes a non-missing factor for the variable)
                # then skip this variable.
                if missing_flag and not eligible:
                    continue

                # Add the linear term to B
                predictor = self.predictors[variable]

#                 # If necessary, protect from missing data
#                 if missing_flag:
#                     B[missing[:, variable]==1, k] = 0.
#                     b = B[:, k]
#                     # Update the outcome data
#                     linear_dependence = self.orthonormal_update(b)

                if missing_flag and not covered:
                    p = B[:, parent_idx] * (1 - missing[:, variable])
                    b = B[:, parent_idx] * (1 - missing[:, variable])
                    self.orthonormal_update(b)
                    b = B[:, parent_idx] * missing[:, variable]
                    self.orthonormal_update(b)
                    q = k + 3
                else:
                    p = self.B[:, parent_idx]
                    q = k + 1

                b = p * predictor.x
                if missing_flag and not covered:
                    b[missing[:, variable] == 1] = 0
                linear_dependence = self.orthonormal_update(b)

                # If a new hinge function does not improve the gcv over the
                # linear term then just the linear term will be retained
                # (if variable_can_be_linear).  Calculate the gcv with just the linear
                # term in order to compare later.  Note that the mse with
                # another term never increases, but the gcv may because it
                # penalizes additional terms.
                mse_ = self.outcome.mse()
                if missing_flag and not covered:
                    gcv_ = gcv_factor_k_plus_3 * mse_
                else:
                    gcv_ = gcv_factor_k_plus_1 * mse_

                if linear_variables[variable]:
                    mse = mse_
                    knot_idx = -1
                else:
                    # Find the valid knot candidates
                    candidates, candidates_idx = predictor.knot_candidates(p, self.endspan,
                                                                           self.minspan,
                                                                           self.minspan_alpha,
                                                                           self.n, set(parent.knots(variable)))
                    # Choose the best candidate (if no candidate is an
                    # improvement on the linear term in terms of gcv, knot_idx
                    # is set to -1
                    if len(candidates_idx) > 0:
#                         candidates = np.array(predictor.x)[candidates_idx]

                        # Find the best knot location for this parent and
                        # variable combination
                        # Assemble the knot search data structure
                        constant = KnotSearchReadOnlyData(predictor, self.outcome)
                        search_data = KnotSearchData(constant, self.workings, q)

                        # Run knot search
                        knot, knot_idx, mse = knot_search(search_data, candidates, p, q,
                                                          self.m, len(candidates), self.n_outcomes,
                                                          self.verbose)
                        mse /= self.total_weight
                        knot_idx = candidates_idx[knot_idx]

                        # If the hinge function does not decrease the gcv then
                        # just keep the linear term (if variable_can_be_linear is True)
                        if variable_can_be_linear:
                            if missing_flag and not covered:
                                if gcv_factor_k_plus_4 * mse >= gcv_:
                                    mse = mse_
                                    knot_idx = -1
                            else:
                                if gcv_factor_k_plus_2 * mse >= gcv_:
                                    mse = mse_
                                    knot_idx = -1
                    else:
                        if variable_can_be_linear:
                            mse = mse_
                            knot_idx = -1
                        else:
                            # Do an orthonormal downdate and skip to the next
                            # iteration
                            if missing_flag and not covered:
                                self.orthonormal_downdate()
                                self.orthonormal_downdate()
                            self.orthonormal_downdate()
                            continue

                # Do an orthonormal downdate
                if missing_flag and not covered:
                    self.orthonormal_downdate()
                    self.orthonormal_downdate()
                self.orthonormal_downdate()

                # Update the choices
                if mse < mse_choice or first:
                    if first:
                        first = False
                    knot_choice = knot
                    mse_choice = mse
                    knot_idx_choice = knot_idx
                    parent_idx_choice = parent_idx
                    parent_choice = parent
                    if self.use_fast is True:
                        parent_basis_content_choice = parent_basis_content
                    variable_choice = variable
                    dependent = linear_dependence
                    if missing_flag and not covered:
                        choice_needs_coverage = True
                    else:
                        choice_needs_coverage = False

                if self.use_fast is True:
                    if (mse_choice_cur_parent == -1) or \
                       (mse < mse_choice_cur_parent):
                        mse_choice_cur_parent = mse
                        variable_choice_cur_parent = variable
            if self.use_fast is True:
                if mse_choice_cur_parent != -1:
                    parent_basis_content.mse = mse_choice_cur_parent
                    parent_basis_content.v = variable_choice_cur_parent

        if self.use_fast is True:
            for content in content_to_be_repushed:
                heappush(self.fast_heap, content)

        # Make sure at least one candidate was checked
        if first:
            self.record[len(self.record) - 1].set_no_candidates(True)
            return

        # Add the new basis functions
        label = self.xlabels[variable_choice]
        if self.use_fast is True:
            parent_basis_content_choice.m = -np.inf
        if choice_needs_coverage:
            new_parent = parent_choice.get_coverage(variable_choice)
            if new_parent is None:
                new_basis_function = MissingnessBasisFunction(parent_choice, variable_choice,
                                               True, label)
                new_basis_function.apply(X, missing, B[:, len(self.basis)])
                self.orthonormal_update(B[:, len(self.basis)])
                if self.use_fast and new_basis_function.is_splittable() and new_basis_function.effective_degree() < self.max_degree:
                    content = FastHeapContent(idx=len(self.basis))
                    heappush(self.fast_heap, content)
                self.basis.append(new_basis_function)
                new_parent = new_basis_function

                new_basis_function = MissingnessBasisFunction(parent_choice, variable_choice,
                                               False, label)
                new_basis_function.apply(X, missing, B[:, len(self.basis)])
                self.orthonormal_update(B[:, len(self.basis)])
                if self.use_fast and new_basis_function.is_splittable() and new_basis_function.effective_degree() < self.max_degree:
                    content = FastHeapContent(idx=len(self.basis))
                    heappush(self.fast_heap, content)
                self.basis.append(new_basis_function)
        else:
            new_parent = parent_choice
        if knot_idx_choice != -1:
            # Add the new basis functions
            new_basis_function = HingeBasisFunction(new_parent,
                                     knot_choice, knot_idx_choice,
                                     variable_choice,
                                     False, label)
            new_basis_function.apply(X, missing, B[:, len(self.basis)])
            self.orthonormal_update(B[:, len(self.basis)])
            if self.use_fast and new_basis_function.is_splittable() and new_basis_function.effective_degree() < self.max_degree:
                content = FastHeapContent(idx=len(self.basis))
                heappush(self.fast_heap, FastHeapContent(idx=len(self.basis)))
            self.basis.append(new_basis_function)

            new_basis_function = HingeBasisFunction(new_parent,
                                     knot_choice, knot_idx_choice,
                                     variable_choice,
                                     True, label)
            new_basis_function.apply(X, missing, B[:, len(self.basis)])
            self.orthonormal_update(B[:, len(self.basis)])
            if self.use_fast and new_basis_function.is_splittable() and new_basis_function.effective_degree() < self.max_degree:
                content = FastHeapContent(idx=len(self.basis))
                heappush(self.fast_heap, content)
            self.basis.append(new_basis_function)

        elif not dependent and knot_idx_choice == -1:
            # In this case, only add the linear basis function (in addition to
            # covering missingness basis functions if needed)
            new_basis_function = LinearBasisFunction(new_parent, variable_choice, label)
            new_basis_function.apply(X, missing, B[:, len(self.basis)])
            self.orthonormal_update(B[:, len(self.basis)])
            if self.use_fast and new_basis_function.is_splittable() and new_basis_function.effective_degree() < self.max_degree:
                content = FastHeapContent(idx=len(self.basis))
                heappush(self.fast_heap, content)
            self.basis.append(new_basis_function)
        else:  # dependent and knot_idx_choice == -1
            # In this case there were no acceptable choices remaining, so end
            # the forward pass
            self.record[len(self.record) - 1].set_no_candidates(True)
            return

        # Compute the new mse, which is the result of the very stable
        # orthonormal updates and not the mse that comes directly from
        # the knot search
        cdef FLOAT_t final_mse = self.outcome.mse()

        # Update the build record
        self.record.append(ForwardPassIteration(parent_idx_choice,
                                                variable_choice,
                                                knot_idx_choice, final_mse,
                                                len(self.basis)))



from __future__ import annotations

import pickle

import numpy

from pyearth._basis import (
    Basis,
    ConstantBasisFunction,
    HingeBasisFunction,
    LinearBasisFunction,
    SmoothedHingeBasisFunction,
)

from . import assert_equal, assert_true
from .base import BaseContainer


class Container(BaseContainer):
    def __init__(self):
        super(Container, self).__init__()
        self.basis = Basis(self.X.shape[1])
        self.parent = ConstantBasisFunction()
        self.bf1 = HingeBasisFunction(self.parent, 1.0, 10, 1, False)
        self.bf2 = HingeBasisFunction(self.parent, 1.0, 4, 2, True)
        self.bf3 = HingeBasisFunction(self.bf2, 1.0, 4, 3, True)
        self.bf4 = LinearBasisFunction(self.parent, 2)
        self.bf5 = HingeBasisFunction(self.parent, 1.5, 8, 2, True)
        self.basis.append(self.parent)
        self.basis.append(self.bf1)
        self.basis.append(self.bf2)
        self.basis.append(self.bf3)
        self.basis.append(self.bf4)
        self.basis.append(self.bf5)


def test_anova_decomp():
    cnt = Container()
    anova = cnt.basis.anova_decomp()
    assert_equal(set(anova[frozenset([1])]), set([cnt.bf1]))
    assert_equal(set(anova[frozenset([2])]), set([cnt.bf2, cnt.bf4, cnt.bf5]))
    assert_equal(set(anova[frozenset([2, 3])]), set([cnt.bf3]))
    assert_equal(set(anova[frozenset()]), set([cnt.parent]))
    assert_equal(len(anova), 4)


def test_smooth_knots():
    cnt = Container()
    mins = [0.0, -1.0, 0.1, 0.2]
    maxes = [2.5, 3.5, 3.0, 2.0]
    knots = cnt.basis.smooth_knots(mins, maxes)
    assert_equal(knots[cnt.bf1], (0.0, 2.25))
    assert_equal(knots[cnt.bf2], (0.55, 1.25))
    assert_equal(knots[cnt.bf3], (0.6, 1.5))
    assert_true(cnt.bf4 not in knots)
    assert_equal(knots[cnt.bf5], (1.25, 2.25))


def test_smooth():
    cnt = Container()
    X = numpy.random.uniform(-2.0, 4.0, size=(20, 4))
    smooth_basis = cnt.basis.smooth(X)
    for bf, smooth_bf in zip(cnt.basis, smooth_basis, strict=False):
        if type(bf) is HingeBasisFunction:
            assert_true(type(smooth_bf) is SmoothedHingeBasisFunction)
        elif type(bf) is ConstantBasisFunction:
            assert_true(type(smooth_bf) is ConstantBasisFunction)
        elif type(bf) is LinearBasisFunction:
            assert_true(type(smooth_bf) is LinearBasisFunction)
        else:
            raise AssertionError("Basis function is of an unexpected type.")
        assert_true(type(smooth_bf) in {SmoothedHingeBasisFunction, ConstantBasisFunction, LinearBasisFunction})
        if bf.has_knot():
            assert_equal(bf.get_knot(), smooth_bf.get_knot())


def test_add():
    cnt = Container()
    assert_equal(len(cnt.basis), 6)


def test_pickle_compat():
    cnt = Container()
    basis_copy = pickle.loads(pickle.dumps(cnt.basis))
    assert_true(cnt.basis == basis_copy)



from __future__ import annotations

import importlib
import importlib.util
import os
import sys

# from distutils.version import LooseVersion
from functools import wraps

import pytest
from numpy.testing import assert_almost_equal


def if_environ_has(var_name):
    # Test decorator that skips test if environment variable is not defined
    def if_environ(func):
        @wraps(func)
        def run_test(*args, **kwargs):
            if var_name in os.environ:
                return func(*args, **kwargs)
            else:
                pytest.skip("Only run if %s environment variable is defined." % var_name)

        return run_test

    return if_environ


def if_platform_not_win_32(func):
    @wraps(func)
    def run_test(*args, **kwargs):
        if sys.platform == "win32":
            pytest.skip("Skip for 32 bit Windows platforms.")
        else:
            return func(*args, **kwargs)

    return run_test


# def if_sklearn_version_greater_than_or_equal_to(min_version):
#     """
#     Test decorator that skips test unless sklearn version is greater than or
#     equal to min_version.
#     """

#     def _if_sklearn_version(func):
#         @wraps(func)
#         def run_test(*args, **kwargs):
#             import sklearn

#             if LooseVersion(sklearn.__version__) < LooseVersion(min_version):
#                 pytest.skip("sklearn version less than %s" % str(min_version))
#             else:
#                 return func(*args, **kwargs)

#         return run_test

#     return _if_sklearn_version


def if_statsmodels(func):
    """Test decorator that skips test if statsmodels not installed."""

    @wraps(func)
    def run_test(*args, **kwargs):
        if importlib.util.find_spec("statsmodels") is None:
            pytest.skip("statsmodels not available.")
        else:
            return func(*args, **kwargs)

    return run_test


def if_pandas(func):
    """Test decorator that skips test if pandas not installed."""

    @wraps(func)
    def run_test(*args, **kwargs):
        if importlib.util.find_spec("pandas") is None:
            pytest.skip("pandas not available.")
        else:
            return func(*args, **kwargs)

    return run_test


def if_sympy(func):
    """Test decorator that skips test if sympy not installed"""

    @wraps(func)
    def run_test(*args, **kwargs):
        if importlib.util.find_spec("sympy") is None:
            pytest.skip("sympy not available.")
        else:
            return func(*args, **kwargs)

    return run_test


def if_patsy(func):
    """Test decorator that skips test if patsy not installed."""

    @wraps(func)
    def run_test(*args, **kwargs):
        if importlib.util.find_spec("patsy") is None:
            pytest.skip("patsy not available.")
        else:
            return func(*args, **kwargs)

    return run_test


def assert_list_almost_equal(list1, list2):
    for el1, el2 in zip(list1, list2, strict=False):
        assert_almost_equal(el1, el2)


def assert_list_almost_equal_value(list, value):
    for el in list:
        assert_almost_equal(el, value)



# distutils: language = c
# cython: cdivision = True
# cython: boundscheck = False
# cython: wraparound = False
# cython: profile = False

from ._util cimport gcv, ascii_table


MAXTERMS = 0
MAXRSQ = 1
NOIMPRV = 2
LOWGRSQ = 3
NOCAND = 4
stopping_conditions = {
    MAXTERMS: "Reached maximum number of terms",
    MAXRSQ: "Achieved RSQ value within threshold of 1",
    NOIMPRV: "Improvement below threshold",
    LOWGRSQ: "GRSQ too low",
    NOCAND: "No remaining candidate knot locations"
}

cdef class Record:

    def __richcmp__(self, other, method):
        if method == 2:
            return self._eq(other)
        elif method == 3:
            return not self._eq(other)
        else:
            return NotImplemented

    def _eq(self, other):
        return (self.__class__ is other.__class__ and
                self._getstate() == other._getstate())

    def __getitem__(Record self, int idx):
        return self.iterations[idx]

    def __len__(Record self):
        return len(self.iterations)

    cpdef append(Record self, Iteration iteration):
        self.iterations.append(iteration)

    cpdef FLOAT_t mse(Record self, INDEX_t iteration):
        return self.iterations[iteration].get_mse()

    cpdef FLOAT_t gcv(Record self, INDEX_t iteration):
        cdef Iteration it = self.iterations[iteration]
        cdef FLOAT_t mse = it.mse
        return gcv(mse, it.get_size(), self.num_samples, self.penalty)

    cpdef FLOAT_t rsq(Record self, INDEX_t iteration):
        # gcv(self.sst,1,self.num_samples,self.penalty)
        cdef FLOAT_t mse0 = self.sst
        # gcv(self.mse(iteration):,self.iterations[iteration].get_size(),
        #        self.num_samples,self.penalty)#self.gcv(iteration)
        cdef FLOAT_t mse = self.mse(iteration)
        return 1 - (mse / mse0)

    cpdef FLOAT_t grsq(Record self, INDEX_t iteration):
        cdef FLOAT_t gcv0 = gcv(self.sst, 1, self.num_samples, self.penalty)
        cdef FLOAT_t gcv_ = self.gcv(iteration)
        return 1 - (gcv_ / gcv0)

cdef class PruningPassRecord(Record):
    def __init__(PruningPassRecord self, INDEX_t num_samples,
                 INDEX_t num_variables, FLOAT_t penalty,
                 FLOAT_t sst, INDEX_t size, FLOAT_t mse):
        self.num_samples = num_samples
        self.num_variables = num_variables
        self.penalty = penalty
        self.sst = sst
        self.iterations = [FirstPruningPassIteration(size, mse)]

    def __reduce__(PruningPassRecord self):
        return (PruningPassRecord, (1, 1, 1.0, 1.0, 1, 1.0), self._getstate())

    def _getstate(PruningPassRecord self):
        result = {'num_samples': self.num_samples,
                  'num_variables': self.num_variables,
                  'penalty': self.penalty,
                  'sst': self.sst,
                  'iterations': self.iterations,
                  'selected': self.selected}
        return result

    def __setstate__(PruningPassRecord self, dict state):
        self.num_samples = state['num_samples']
        self.num_variables = state['num_variables']
        self.penalty = state['penalty']
        self.sst = state['sst']
        self.iterations = state['iterations']
        self.selected = state['selected']

    cpdef set_selected(PruningPassRecord self, INDEX_t selected):
        self.selected = selected

    cpdef INDEX_t get_selected(PruningPassRecord self):
        return self.selected

    cpdef roll_back(PruningPassRecord self, Basis basis):
        cdef INDEX_t n = len(self.iterations)
        cdef INDEX_t i
        for i in range(n - self.selected - 1):
            basis[self.iterations[n - i - 1].get_pruned()].unprune()

    def __str__(PruningPassRecord self):
        return self.partial_str(slice(None))

    def partial_str(PruningPassRecord self, rows, print_header=True, print_footer=True):
        result = ''
#         result += 'Pruning Pass\n'
        header = 'iter\tbf\tterms\tmse\tgcv\trsq\tgrsq'.split('\t')
        data = []
        map_back = list(range(*rows.indices(len(self.iterations))))
        for i, iteration in enumerate(self.iterations[rows]):
            row = str(map_back[i]) + '\t' + str(iteration) + '\t%.3f\t%.3f\t%.3f' % (
                self.gcv(map_back[i]), self.rsq(map_back[i]), self.grsq(map_back[i]))
            data.append(row.split('\t'))
        result += ascii_table(header, data, print_header, print_footer)
#         result += '\nSelected iteration: ' + str(self.selected) + '\n'
        return result

    def final_str(PruningPassRecord self):
        return 'Selected iteration: ' + str(self.selected)

cdef class ForwardPassRecord(Record):
    def __init__(ForwardPassRecord self,
                 INDEX_t num_samples, INDEX_t num_variables,
                 FLOAT_t penalty, FLOAT_t sst,
                 list xlabels):
        self.num_samples = num_samples
        self.num_variables = num_variables
        self.penalty = penalty
        self.sst = sst
        self.iterations = [FirstForwardPassIteration(self.sst)]
        self.xlabels = xlabels

    def __reduce__(ForwardPassRecord self):
        return (ForwardPassRecord, (self.num_samples, self.num_variables,
                                    self.penalty, self.sst, self.xlabels),
                self._getstate())

    def _getstate(ForwardPassRecord self):
        return {'num_samples': self.num_samples,
                'num_variables': self.num_variables,
                'penalty': self.penalty,
                'sst': self.sst,
                'iterations': self.iterations,
                'xlabels': self.xlabels}

    def __setstate__(ForwardPassRecord self, dict state):
        self.num_samples = state['num_samples']
        self.num_variables = state['num_variables']
        self.penalty = state['penalty']
        self.sst = state['sst']
        self.iterations = state['iterations']
        self.xlabels = state['xlabels']

    cpdef set_stopping_condition(ForwardPassRecord self,
                                 int stopping_condition):
        self.stopping_condition = stopping_condition

    def __str__(ForwardPassRecord self):
        return self.partial_str(slice(None))
#         header = ['iter', 'parent', 'var', 'knot',
#                   'mse', 'terms', 'gcv', 'rsq', 'grsq']
#         data = []
#         for i, iteration in enumerate(self.iterations):
#             data.append([str(i)] + str(iteration).split('\t') +
#                         ('%.3f\t%.3f\t%.3f' % (self.gcv(i), self.rsq(i),
#                                                self.grsq(i))).split('\t'))
#         result = ''
#         result += 'Forward Pass\n'
#         result += ascii_table(header, data)
#         result += '\nStopping Condition %d: %s\n' % (
#             self.stopping_condition,
#             stopping_conditions[self.stopping_condition])
#         return result

    def partial_str(ForwardPassRecord self, rows, print_header=True, print_footer=True):
        header = ['iter', 'parent', 'var', 'knot',
                  'mse', 'terms', 'gcv', 'rsq', 'grsq']
        data = []
        map_back = list(range(*rows.indices(len(self.iterations))))
        for i, iteration in enumerate(self.iterations[rows]):
            data.append([str(map_back[i])] + str(iteration).split('\t') +
                        ('%.3f\t%.3f\t%.3f' % (self.gcv(map_back[i]), self.rsq(map_back[i]),
                                               self.grsq(map_back[i]))).split('\t'))
        result = ''
#         result += 'Forward Pass\n'
        result += ascii_table(header, data, print_header, print_footer)
#         result += '\nStopping Condition %d: %s\n' % (
#             self.stopping_condition,
#             stopping_conditions[self.stopping_condition])
        return result

    def final_str(ForwardPassRecord self):
        return 'Stopping Condition %d: %s' % (
            self.stopping_condition,
            stopping_conditions[self.stopping_condition])

cdef class Iteration:

    def __richcmp__(self, other, method):
        if method == 2:
            return self._eq(other)
        elif method == 3:
            return not self._eq(other)
        else:
            return NotImplemented

    def _eq(self, other):
        return (self.__class__ is other.__class__ and
                self._getstate() == other._getstate())

    cpdef FLOAT_t get_mse(Iteration self):
        return self.mse

    cpdef INDEX_t get_size(Iteration self):
        return self.size

cdef class PruningPassIteration(Iteration):
    def __init__(PruningPassIteration self,
                 INDEX_t pruned, INDEX_t size,
                 FLOAT_t mse):
        self.pruned = pruned
        self.size = size
        self.mse = mse

    def __reduce__(PruningPassIteration self):
        return (PruningPassIteration, (1, 1, 1.0), self._getstate())

    def _getstate(PruningPassIteration self):
        return {'pruned': self.pruned,
                'size': self.size,
                'mse': self.mse}

    def __setstate__(PruningPassIteration self, dict state):
        self.pruned = state['pruned']
        self.size = state['size']
        self.mse = state['mse']

    cpdef INDEX_t get_pruned(PruningPassIteration self):
        return self.pruned

    def __str__(PruningPassIteration self):
        result = '%s\t%s\t%s' % (str(self.pruned), self.size, '%.2f' %
                                 self.mse if self.mse is not None else None)
        return result

cdef class FirstPruningPassIteration(PruningPassIteration):
    def __init__(PruningPassIteration self, INDEX_t size, FLOAT_t mse):
        self.size = size
        self.mse = mse

    def __reduce__(FirstPruningPassIteration self):
        return (FirstPruningPassIteration, (1, 1.0), self._getstate())

    def _getstate(FirstPruningPassIteration self):
        return {'size': self.size,
                'mse': self.mse}

    def __setstate__(FirstPruningPassIteration self, dict state):
        self.size = state['size']
        self.mse = state['mse']

    def __str__(PruningPassIteration self):
        result = '%s\t%s\t%s' % ('-', self.size, '%.2f' %
                                 self.mse if self.mse is not None else None)
        return result

cdef class ForwardPassIteration(Iteration):
    def __init__(ForwardPassIteration self,
                 INDEX_t parent, INDEX_t variable,
                 int knot, FLOAT_t mse,
                 INDEX_t size):
        self.parent = parent
        self.variable = variable
        self.knot = knot
        self.mse = mse
        self.size = size

    def __reduce__(ForwardPassIteration self):
        return (ForwardPassIteration, (1, 1, 1, 1.0, 1), self._getstate())

    def _getstate(ForwardPassIteration self):
        return {'parent': self.parent,
                'variable': self.variable,
                'knot': self.knot,
                'mse': self.mse,
                'size': self.size}

    def __setstate__(ForwardPassIteration self, dict state):
        self.parent = state['parent']
        self.variable = state['variable']
        self.knot = state['knot']
        self.mse = state['mse']
        self.size = state['size']

    def __str__(self):
        result = '%d\t%d\t%d\t%4f\t%d' % (
            self.parent, self.variable, self.knot, self.mse, self.size)
        return result

    cpdef set_no_candidates(ForwardPassIteration self, bint value):
        self.no_candidates = value

    cpdef no_further_candidates(ForwardPassIteration self):
        return self.no_candidates

cdef class FirstForwardPassIteration(ForwardPassIteration):
    def __init__(FirstForwardPassIteration self, FLOAT_t mse):
        self.mse = mse

    def __reduce__(FirstForwardPassIteration self):
        return (FirstForwardPassIteration, (1.0,), self._getstate())

    def _getstate(FirstForwardPassIteration self):
        return {'mse': self.mse}

    def __setstate__(FirstForwardPassIteration self, dict state):
        self.mse = state['mse']

    cpdef INDEX_t get_size(FirstForwardPassIteration self):
        return 1

    def __str__(self):
        result = '%s\t%s\t%s\t%4f\t%s' % ('-', '-', '-', self.mse, 1)
        return result



from __future__ import annotations

from pyearth._record import ForwardPassIteration, ForwardPassRecord
from pyearth._util import gcv

from ..testing_utils import assert_list_almost_equal

num_samples = 1000
num_variables = 10
penalty = 3.0
sst = 100.0
varnames = ["x" + str(i) for i in range(num_variables)]
record = ForwardPassRecord(num_samples, num_variables, penalty, sst, varnames)
record.append(ForwardPassIteration(0, 3, 3, 63.0, 3))
record.append(ForwardPassIteration(0, 3, 14, 34.0, 5))
record.append(ForwardPassIteration(3, 6, 12, 18.0, 7))
mses = [sst, 63.0, 34.0, 18.0]
sizes = [1, 3, 5, 7]


def test_statistics():
    mses = [record.mse(i) for i in range(len(record))]
    mses_ = [mses[i] for i in range(len(record))]
    gcvs = [record.gcv(i) for i in range(len(record))]
    gcvs_ = [gcv(mses[i], sizes[i], num_samples, penalty) for i in range(len(record))]
    rsqs = [record.rsq(i) for i in range(len(record))]
    rsqs_ = [1 - (mses[i] / sst) for i in range(len(record))]
    grsqs = [record.grsq(i) for i in range(len(record))]
    grsqs_ = [1 - (record.gcv(i) / gcv(sst, 1, num_samples, penalty)) for i in range(len(record))]
    assert_list_almost_equal(mses, mses_)
    assert_list_almost_equal(gcvs, gcvs_)
    assert_list_almost_equal(rsqs, rsqs_)
    assert_list_almost_equal(grsqs, grsqs_)



"""
============================================
Plotting derivatives of simple sine function
============================================

A simple example plotting a fit of the sine function and
the derivatives computed by Earth.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy

from pyearth import Earth

# Create some fake data
numpy.random.seed(2)
m = 10000
n = 10
X = 20 * numpy.random.uniform(size=(m, n)) - 10
y = 10 * numpy.sin(X[:, 6]) + 0.25 * numpy.random.normal(size=m)

# Compute the known true derivative with respect to the predictive variable
y_prime = 10 * numpy.cos(X[:, 6])

# Fit an Earth model
model = Earth(max_degree=2, minspan_alpha=0.5, smooth=True)
model.fit(X, y)

# Print the model
print(model.trace())
print(model.summary())

# Get the predicted values and derivatives
y_hat = model.predict(X)
y_prime_hat = model.predict_deriv(X, "x6")

# Plot true and predicted function values and derivatives
# for the predictive variable
plt.subplot(211)
plt.plot(X[:, 6], y, "r.")
plt.plot(X[:, 6], y_hat, "b.")
plt.ylabel("function")
plt.subplot(212)
plt.plot(X[:, 6], y_prime, "r.")
plt.plot(X[:, 6], y_prime_hat[:, 0], "b.")
plt.ylabel("derivative")
plt.show()



from __future__ import annotations

import pickle

import numpy

from pyearth._basis import ConstantBasisFunction
from pyearth._types import BOOL

from . import assert_false, assert_true
from .base import BaseContainer


class Container(BaseContainer):
    def __init__(self):
        super(Container, self).__init__()
        self.bf = ConstantBasisFunction()


def test_apply():
    cnt = Container()
    m, _ = cnt.X.shape
    missing = numpy.zeros_like(cnt.X, dtype=BOOL)
    B = numpy.empty(shape=(m, 10))
    assert_false(numpy.all(B[:, 0] == 1))
    cnt.bf.apply(cnt.X, missing, B[:, 0])
    assert_true(numpy.all(B[:, 0] == 1))


def test_deriv():
    cnt = Container()
    m, _ = cnt.X.shape
    missing = numpy.zeros_like(cnt.X, dtype=BOOL)
    b = numpy.empty(shape=m)
    j = numpy.empty(shape=m)
    cnt.bf.apply_deriv(cnt.X, missing, b, j, 1)
    assert_true(numpy.all(b == 1))
    assert_true(numpy.all(j == 0))


def test_pickle_compatibility():
    cnt = Container()
    bf_copy = pickle.loads(pickle.dumps(cnt.bf))
    assert_true(cnt.bf == bf_copy)


def test_smoothed_version():
    cnt = Container()
    smoothed = cnt.bf._smoothed_version(None, {}, {})
    assert_true(type(smoothed) is ConstantBasisFunction)



from __future__ import annotations

from timeit import Timer

import numpy as np

from pyearth import Earth

# The robot arm example, as defined in:
# Fast MARS, Jerome H.Friedman, Technical Report No.110, May 1993, section 6.2.

np.random.seed(2)
nb_examples = 400
theta1 = np.random.uniform(0, 2 * np.pi, size=nb_examples)
theta2 = np.random.uniform(0, 2 * np.pi, size=nb_examples)
phi = np.random.uniform(-np.pi / 2, np.pi / 2, size=nb_examples)
l1 = np.random.uniform(0, 1, size=nb_examples)
l2 = np.random.uniform(0, 1, size=nb_examples)
x = l1 * np.cos(theta1) - l2 * np.cos(theta1 + theta2) * np.cos(phi)
y = l1 * np.sin(theta1) - l2 * np.sin(theta1 + theta2) * np.cos(phi)
z = l2 * np.sin(theta2) * np.sin(phi)
d = np.sqrt(x**2 + y**2 + z**2)


inputs = np.concatenate(
    [
        theta1[:, np.newaxis],
        theta2[:, np.newaxis],
        phi[:, np.newaxis],
        l1[:, np.newaxis],
        l2[:, np.newaxis],
    ],
    axis=1,
)
outputs = d

hp = dict(
    max_degree=5,
    minspan=1,
    endspan=1,
    max_terms=100,
    allow_linear=False,
)
model_normal = Earth(**hp)
t = Timer(lambda: model_normal.fit(inputs, outputs))
duration_normal = t.timeit(number=1)
print("Normal : MSE={0:.5f}, duration={1:.2f}s".format(model_normal.mse_, duration_normal))
model_fast = Earth(use_fast=True, fast_K=5, fast_h=1, **hp)

t = Timer(lambda: model_fast.fit(inputs, outputs))
duration_fast = t.timeit(number=1)
print("Fast: MSE={0:.5f}, duration={1:.2f}s".format(model_fast.mse_, duration_fast))
speedup = duration_normal / duration_fast
print(
    "diagnostic : MSE goes from {0:.5f} to {1:.5f} but it is {2:.2f}x faster".format(
        model_normal.mse_, model_fast.mse_, speedup
    )
)



from __future__ import annotations


def assert_equal(x, y):
    assert x == y


def assert_not_equal(x, y):
    assert x != y


def assert_true(x):
    assert x


def assert_false(x):
    assert not x



# distutils: language = c
# cython: cdivision = True
# cython: boundscheck = False
# cython: wraparound = False
# cython: profile = True
cimport cython
import numpy as np
import scipy as sp
from libc.math cimport sqrt
from libc.math cimport log
cimport numpy as cnp
from ._types import INDEX, FLOAT
from ._util cimport log2


@cython.final
cdef class SingleWeightDependentData:
    def __init__(SingleWeightDependentData self, UpdatingQT updating_qt, FLOAT_t[:] w, INDEX_t m,
                 INDEX_t k, INDEX_t max_terms, FLOAT_t total_weight):
        self.updating_qt = updating_qt
        self.w = w
        self.m = m
        self.k = k
        self.max_terms = max_terms
        self.Q_t = self.updating_qt.Q_t
        self.total_weight = total_weight

    @classmethod
    def alloc(cls, FLOAT_t[:] w, INDEX_t m, INDEX_t max_terms, FLOAT_t zero_tol):
        cdef UpdatingQT updating_qt = UpdatingQT.alloc(m, max_terms, zero_tol)
        cdef FLOAT_t total_weight = np.dot(w[:m], w[:m])
        return cls(updating_qt, w, m, 0, max_terms, total_weight)

#     cpdef int update_from_basis_function(SingleWeightDependentData self, BasisFunction bf, FLOAT_t[:,:] X,
#                                          BOOL_t[:,:] missing) except *:
#         if self.k >= self.max_terms:
#             return -1
#         bf.apply(X, missing, self.Q_t[self.k, :])
#         return self._update()

    cpdef int update_from_array(SingleWeightDependentData self, FLOAT_t[:] b) except *:
        if self.k >= self.max_terms:
            return -1

        self.updating_qt.update(np.asarray(self.w) * b)
        self.k += 1
#         cdef INDEX_t j
#         for j in range(self.m):
#             self.Q_t[self.k,j] = self.w[j] * b[j]
#         return self._update(zero_tol)
#
#     cpdef int _update(SingleWeightDependentData self, FLOAT_t zero_tol):
#         # Compute the new householder reflection
#         np.asarray(self.Q_t)[self.k, :] = self.householder.apply_transpose(self.Q_t[self.k, :])
#         self.householder.push_from_column(self.Q_t[self.k, self.k], self.Q_t[self.k,(self.k + 1):])
#
#         # Create the new row in Q_t and apply all existing householder reflections
#         # including the new one
#         self.Q_t[self.k, :] = 0.
#         self.Q_t[self.k, self.k] = 1.
#         np.asarray(self.Q_t)[self.k, :] = self.householder.apply(self.Q_t[self.k, :])
#
#         self.k += 1

    cpdef downdate(SingleWeightDependentData self):
        self.updating_qt.downdate()
        self.k -= 1

    cpdef reweight(SingleWeightDependentData self, FLOAT_t[:] w, FLOAT_t[:,:] B, INDEX_t k):
        cdef INDEX_t i
        self.w = w
        self.total_weight = np.dot(self.w[:self.m], self.w[:self.m])
        self.k = 0
        self.updating_qt.reset()
        for i in range(k):
            self.update_from_array(B[:, i])

@cython.final
cdef class MultipleOutcomeDependentData:
    def __init__(MultipleOutcomeDependentData self, list outcomes, list weights):
        self.outcomes = outcomes
        self.weights = weights

    @classmethod
    def alloc(cls, FLOAT_t[:,:] y, w, INDEX_t m, INDEX_t n_outcomes, INDEX_t max_terms, FLOAT_t zero_tol):
        cdef list weights
        cdef list outcomes
        cdef int i, n_weights
        # w is a numpy array of weights
        if len(w.shape) == 2 and w.shape[1] == n_outcomes:
            n_weights = w.shape[1]
            weights = []
            for i in range(w.shape[1]):
                weights.append(SingleWeightDependentData.alloc(w[:, i], m, max_terms, zero_tol))
        elif len(w.shape) == 1 or w.shape[1] == 1:
            n_weights = 1
            if len(w.shape) == 1:
                weights = [SingleWeightDependentData.alloc(w, m, max_terms, zero_tol)]
            else:
                weights = [SingleWeightDependentData.alloc(w[:, 0], m, max_terms, zero_tol)]
        else:
            raise ValueError('Shape of weights does not match shape of outcomes.')

        outcomes = []
        for i in range(n_outcomes):
            outcomes.append(SingleOutcomeDependentData.alloc(y[:, i], weights[i % n_weights], m, max_terms))

        return cls(outcomes, weights)

    cpdef update_from_array(MultipleOutcomeDependentData self, FLOAT_t[:] b):
        cdef SingleWeightDependentData weight
        cdef SingleOutcomeDependentData outcome
        for weight in self.weights:
            weight.update_from_array(b)
        for outcome in self.outcomes:
            outcome.update()

    cpdef downdate(MultipleOutcomeDependentData self):
        cdef SingleWeightDependentData weight
        cdef SingleOutcomeDependentData outcome
        for weight in self.weights:
            weight.downdate()
        for outcome in self.outcomes:
            outcome.downdate()

    cpdef list sse(MultipleOutcomeDependentData self):
        return [outcome.sse() for outcome in self.outcomes]

    cpdef FLOAT_t mse(MultipleOutcomeDependentData self):
        cdef FLOAT_t numerator = 0.
        cdef FLOAT_t denominator = 0.
        for outcome in self.outcomes:
            numerator += outcome.sse_
            denominator += outcome.weight.total_weight
        return numerator / denominator
#         return [outcome.sse_ / outcome.weight.total_weight for outcome in self.outcomes]

@cython.final
cdef class SingleOutcomeDependentData:
    def __init__(SingleOutcomeDependentData self, FLOAT_t[:] y, SingleWeightDependentData weight,
                 FLOAT_t[:] theta, FLOAT_t omega, INDEX_t m, INDEX_t k, INDEX_t max_terms, FLOAT_t sse_):
        self.y = y
        self.weight = weight
        self.theta = theta
        self.omega = omega
        self.m = m
        self.k = k
        self.max_terms = max_terms
        self.sse_ = sse_

    @classmethod
    def alloc(cls, FLOAT_t[:] y, SingleWeightDependentData weight, INDEX_t m, INDEX_t max_terms):
        cdef FLOAT_t[:] theta
        cdef FLOAT_t[:] wy = np.empty(shape=m, dtype=np.float64)
        cdef int i
        for i in range(m):
            wy[i] = weight.w[i] * y[i]
        cdef FLOAT_t omega = np.dot(wy, wy)
        cdef FLOAT_t sse_ = omega
        theta = np.dot(weight.Q_t, wy)
        return cls(y, weight, theta, omega, m, 0, max_terms, sse_)

    cpdef FLOAT_t sse(SingleOutcomeDependentData self):
        '''
        Return the weighted mean squared error for the linear least squares problem
        represented by Q_t, y, and w.
        '''
        # TODO: Why is this squared?
        return ((self.omega - np.dot(self.theta, self.theta)) ** 2)# / np.sum(self.w)

#     cpdef int update_from_basis_function(OutcomeDependentData self, BasisFunction bf, FLOAT_t[:,:] X,
#                                          BOOL_t[:,:] missing, FLOAT_t zero_tol) except *:
#         if self.k >= self.max_terms:
#             return -1
#         bf.apply(X, missing, self.Q_t[self.k, :])
#         return self._update(zero_tol)
#
#     cpdef int update_from_array(OutcomeDependentData self, FLOAT_t[:] b, FLOAT_t zero_tol) except *:
#         if self.k >= self.max_terms:
#             return -1
#
#
#         cdef INDEX_t j
#         for j in range(self.m):
#             self.Q_t[self.k,j] = self.w[j] * b[j]
#         return self._update(zero_tol)
    cpdef int synchronize(SingleOutcomeDependentData self) except *:
        self.k = self.weight.k
        self.theta = np.dot(self.weight.Q_t[:self.k, :], np.asarray(self.y) * self.weight.w)
        return 0

    cpdef int update(SingleOutcomeDependentData self) except *:
        # Assume weight has already been updated.
        if self.k >= self.max_terms:
            return -1
        self.k += 1
        self.theta = np.dot(self.weight.Q_t[:self.k, :], np.asarray(self.y) * self.weight.w)
        self.sse_ = self.omega - np.dot(self.theta[:self.k], self.theta[:self.k])

        return 0

    cpdef downdate(SingleOutcomeDependentData self):
        self.k -= 1
        self.sse_ = self.omega - np.dot(self.theta[:self.k], self.theta[:self.k])

#     cpdef reweight(OutcomeDependentData self, FLOAT_t[:] w, FLOAT_t[:,:] B, INDEX_t k, FLOAT_t zero_tol):
#         cdef INDEX_t i
#         for weight in self.weights:
#
#         self.w = w
#         self.k = 0
#         self.householder.reset()
#         for i in range(k):
#             self.update_from_array(B[:, i], zero_tol)


@cython.final
cdef class PredictorDependentData:
    def __init__(PredictorDependentData self, FLOAT_t[:] x,
                INDEX_t[:] order):
        self.x = x
        self.order = order

    def knot_candidates(PredictorDependentData self, cnp.ndarray[FLOAT_t, ndim = 1] p, int endspan,
                        int minspan, FLOAT_t minspan_alpha, INDEX_t n, set knot_set):
        cdef INDEX_t minspan_, i, count, m, idx, countdown
        cdef FLOAT_t last, knot
        cdef bint first, skip
        cdef list candidates = []
        cdef list candidates_idx = []
        m = p.shape[0]
        count = 0
        for i in range(m):
            if p[i] != 0:
                count += 1

        if n * count == 0:
            return np.array(candidates, dtype=FLOAT), np.array(candidates_idx, dtype=INDEX)

        if minspan < 0:
            minspan_ =  <int> (-log2(-(1.0 / (n * count)) *
                                log(1.0 - minspan_alpha)) / 2.5)
        else:
            minspan_ = minspan

        i = endspan
        first = True
        skip = False
        countdown = 0
        while True:
            if m < endspan + i:
                break
            idx = self.order[i]
            knot = self.x[idx]
            if ((not first) and knot == last) or p[idx] == 0 or knot in knot_set:
                countdown = minspan_
                skip = True
                i += 1
            else:
                if first or knot != last:
                    last = knot
                    if countdown <= 0:
                        candidates.append(knot)
                        candidates_idx.append(idx)
                        countdown = minspan_
                    else:
                        countdown -= 1
                i += 1
            first = False

        return np.array(candidates, dtype=FLOAT), np.array(candidates_idx, dtype=INDEX)

    def ordered(self):
        return np.array(self.x)[self.order]

    @classmethod
    def alloc(cls, FLOAT_t[:] x):
        cdef INDEX_t[:] order
        order = np.argsort(x)[::-1]
        return cls(x, order)

@cython.final
cdef class KnotSearchReadOnlyData:
    def __init__(KnotSearchReadOnlyData self, PredictorDependentData predictor, MultipleOutcomeDependentData outcome):
        self.predictor = predictor
        self.outcome = outcome

#     @classmethod
#     def alloc(cls, FLOAT_t[:,:] Q_t, FLOAT_t[:] p, FLOAT_t[:] x,
#               INDEX_t[:] order, FLOAT_t[:] y,
#               FLOAT_t[:] w, int max_terms):
#         cdef int n_outcomes = y.shape[1]
#         cdef PredictorDependentData predictor = PredictorDependentData(p, x,
#                                                         order)
#         cdef list outcomes = []
#         cdef int i
#         cdef MultipleOutcomeDependentDataoutcome = MultipleOutcomeDependentData.alloc():
#         for i in range(n_outcomes):
#             outcomes.append(OutcomeDependentData.alloc(y, w, max_terms))
#         return cls(predictor, outcomes)


@cython.final
cdef class KnotSearchState:
#     FLOAT_t alpha
#     FLOAT_t beta
#     FLOAT_t lambda_
#     FLOAT_t mu
#     FLOAT_t upsilon
#     FLOAT_t phi
#     FLOAT_t phi_next
#     INDEX_t ord_idx
#     INDEX_t idx
#     FLOAT_t zeta_squared
    def __init__(KnotSearchState self, FLOAT_t alpha, FLOAT_t beta, FLOAT_t lambda_,
                 FLOAT_t mu, FLOAT_t upsilon, FLOAT_t phi, FLOAT_t phi_next,
                 INDEX_t ord_idx, INDEX_t idx, FLOAT_t zeta_squared):
        self.alpha = alpha
        self.beta = beta
        self.lambda_ = lambda_
        self.mu = mu
        self.upsilon = upsilon
        self.phi = phi
        self.phi_next = phi_next
        self.ord_idx = ord_idx
        self.idx = idx
        self.zeta_squared = zeta_squared

    @classmethod
    def alloc(cls):
        return cls(0., 0., 0., 0., 0., 0., 0., 0, 0, 0.)

@cython.final
cdef class KnotSearchWorkingData:
    def __init__(KnotSearchWorkingData self, FLOAT_t[:] gamma, FLOAT_t[:] kappa,
                 FLOAT_t[:] delta_kappa, FLOAT_t[:] chi, FLOAT_t[:] psi,
                 KnotSearchState state):
        self.gamma = gamma
        self.kappa = kappa
        self.delta_kappa = delta_kappa
        self.chi = chi
        self.psi = psi
        self.state = state

    @classmethod
    def alloc(cls, int max_terms):
        cdef FLOAT_t[:] gamma = np.empty(shape=max_terms, dtype=np.float64)
        cdef FLOAT_t[:] kappa = np.empty(shape=max_terms, dtype=np.float64)
        cdef FLOAT_t[:] delta_kappa = np.empty(shape=max_terms, dtype=np.float64)
        cdef FLOAT_t[:] chi = np.empty(shape=max_terms, dtype=np.float64)
        cdef FLOAT_t[:] psi = np.empty(shape=max_terms, dtype=np.float64)
        cdef INDEX_t q = 0
        cdef KnotSearchState state = KnotSearchState.alloc()
        return cls(gamma, kappa, delta_kappa, chi, psi, state)

@cython.final
cdef class KnotSearchData:
    def __init__(KnotSearchData self, KnotSearchReadOnlyData constant,
                 list workings, INDEX_t q):
        self.constant = constant
        self.workings = workings
        self.q = q

cdef dot(FLOAT_t[:] x1, FLOAT_t[:] x2, INDEX_t q):
    cdef FLOAT_t result = 0.
    cdef INDEX_t i
    for i in range(q):
        result += x1[i] * x2[i]
    return result

cdef w2dot(FLOAT_t[:] w, FLOAT_t[:] x1, FLOAT_t[:] x2, INDEX_t q):
    cdef FLOAT_t result = 0.
    cdef INDEX_t i
    for i in range(q):
        result += (w[i] ** 2) * x1[i] * x2[i]
    return result

cdef wdot(FLOAT_t[:] w, FLOAT_t[:] x1, FLOAT_t[:] x2, INDEX_t q):
    cdef FLOAT_t result = 0.
    cdef INDEX_t i
    for i in range(q):
        result += w[i] * x1[i] * x2[i]
    return result

@cython.profile(False)
cdef inline void fast_update(PredictorDependentData predictor, SingleOutcomeDependentData outcome,
                        KnotSearchWorkingData working, FLOAT_t[:] p, INDEX_t q, INDEX_t m, INDEX_t r) except *:

    # Calculate all quantities depending on the rows such that
    # phi >= x > phi_next.
    # Before this while loop, x[idx] is the greatest x such that x <= phi.
    # This while loop computes the quantities nu, xi, rho, sigma, tau,
    # chi, and psi.  It also computes the updates to kappa, lambda, mu,
    # and upsilon.  The latter updates should not be applied until after
    # alpha, beta, and gamma have been updated, as they apply to the
    # next iteration.
    cdef FLOAT_t epsilon_squared
    cdef INDEX_t idx, j
    cdef FLOAT_t nu = 0.
    cdef FLOAT_t xi = 0.
    cdef FLOAT_t rho = 0.
    cdef FLOAT_t sigma = 0.
    cdef FLOAT_t tau = 0.
    working.chi[:q] = 0.
    working.psi[:q] = 0.
    working.delta_kappa[:q] = 0.
    cdef FLOAT_t delta_lambda = 0.
    cdef FLOAT_t delta_mu = 0.
    cdef FLOAT_t delta_upsilon = 0.
    cdef FLOAT_t gamma_squared
    cdef FLOAT_t theta_gamma
    cdef FLOAT_t zeta_epsilon
    cdef FLOAT_t tol = .9
    cdef FLOAT_t pidx, xidx, widx, yidx, qidx, delta_nu, \
        delta_xi, delta_rho, delta_sigma, delta_tau, delta_psi, delta_chi

    while predictor.x[working.state.idx] > working.state.phi_next:
        idx = working.state.idx
        pidx = p[idx]
        xidx = predictor.x[idx]
        widx = outcome.weight.w[idx]
        yidx = outcome.y[idx]

        # In predictor.x[idx] is missing, p[idx] will be zeroed out for protection
        # (because there will be a present(x[idx]) factor in it)..
        # Skipping such indices prevents problems if x[idx] is a nan of some kind.
        if p[idx] != 0.:
            delta_nu = (widx ** 2) * (pidx ** 2)
            nu += delta_nu
            delta_xi = delta_nu * xidx
            xi += delta_xi # (outcome.weight.w[idx] ** 2) * (p[idx] ** 2) * predictor.x[idx]
            delta_rho = delta_xi * xidx
            rho += delta_rho # (outcome.weight.w[idx] ** 2) * (p[idx] ** 2) * (predictor.x[idx] ** 2)
            delta_tau = (widx ** 2) * yidx * pidx
            tau += delta_tau
            delta_sigma = delta_tau * xidx
            sigma += delta_sigma # (outcome.weight.w[idx] ** 2) * outcome.y[idx] * p[idx] * predictor.x[idx]
            delta_lambda += delta_xi # (outcome.weight.w[idx] ** 2) * (p[idx] ** 2) * predictor.x[idx]
            delta_mu += delta_nu #(outcome.weight.w[idx] ** 2) * (p[idx] ** 2)
            delta_upsilon += delta_tau # (outcome.weight.w[idx] ** 2) * outcome.y[idx] * p[idx]
            for j in range(q):
                qidx = outcome.weight.Q_t[j,idx]
                delta_psi = qidx * widx * pidx
                delta_chi = delta_psi * xidx
                working.chi[j] += delta_chi # outcome.weight.Q_t[j,idx] * widx * pidx * xidx
                working.psi[j] += delta_psi # outcome.weight.Q_t[j,idx] * outcome.weight.w[idx] * p[idx]
                working.delta_kappa[j] += delta_psi # outcome.weight.Q_t[j,idx] * outcome.weight.w[idx] * p[idx]

        # Update idx for next iteration
        working.state.ord_idx += 1
        if working.state.ord_idx >= m:
            break
        working.state.idx = predictor.order[working.state.ord_idx]

    # Update alpha, beta, and gamma
    working.state.alpha += sigma - working.state.phi_next * tau + \
        (working.state.phi - working.state.phi_next) * working.state.upsilon
    working.state.beta += rho + (working.state.phi_next ** 2) * nu - 2 * working.state.phi_next * xi + \
        2 * (working.state.phi - working.state.phi_next) * working.state.lambda_ + \
        (working.state.phi_next ** 2 - working.state.phi ** 2) * working.state.mu
    for j in range(q):
        working.gamma[j] += (working.state.phi - working.state.phi_next) * working.kappa[j] + \
                            working.chi[j] - working.state.phi_next * working.psi[j]

#     x_should_be = np.maximum(np.asarray(predictor.x) - working.state.phi_next, 0) * p
#     alpha_should_be = np.dot(x_should_be * outcome.weight.w, np.array(outcome.weight.w) * outcome.y)
#     print 'alpha = ', np.asarray(working.state.alpha), alpha_should_be
#     print 'beta =', np.asarray(working.state.beta), np.dot(x_should_be, x_should_be)
#     print 'gamma =', np.asarray(working.gamma[:q]), np.dot(outcome.weight.Q_t[:q,:], x_should_be)

    # Compute epsilon_squared and zeta_squared
    if working.state.beta > 0:
        gamma_squared = dot(working.gamma, working.gamma, q)
        epsilon_squared = working.state.beta - gamma_squared
        if epsilon_squared > 0:
            theta_gamma = dot(working.gamma, outcome.theta, q)
            zeta_epsilon = working.state.alpha - theta_gamma
#             if (abs(zeta_epsilon) / abs(working.state.alpha - theta_gamma) > tol) \
#                 or (epsilon_squared / abs(working.state.beta + gamma_squared) > tol):
            working.state.zeta_squared = (zeta_epsilon ** 2) / epsilon_squared
#             working.state.zeta_squared /= epsilon_squared
#             if epsilon_squared < 1e-6:
#                 print 'epsilon_squared =', epsilon_squared
#                 print 'alpha =', working.state.alpha
#                 print 'gamma * theta =', dot(working.gamma, outcome.theta, q)
#                 print 'beta =', working.state.beta
#             else:
#                 # Numerical instability got us here.  Assume linear
#                 # dependence (which is what causes the instability) and
#                 # set zeta_squared, alpha, and beta accordingly
#                 print 'else'
#                 working.state.zeta_squared = 0.
#                 working.state.beta = gamma_squared
#                 theta_gamma = dot(working.gamma, outcome.theta, q)
#                 working.state.alpha = theta_gamma
        else:
            working.state.zeta_squared = 0.
    else:
        # This happens when there are no nonzero values in the
        # new predictor yet.  It just means we need to wait for
        # lower knot values.
        working.state.zeta_squared = 0.
#         gamma_squared = dot(working.gamma, working.gamma, q)
#         working.state.beta = gamma_squared
#         theta_gamma = dot(working.gamma, outcome.theta, q)
#         working.state.alpha = theta_gamma
    # Now zeta_squared is correct for phi_next.

    # Update kappa, lambda, mu, and upsilon
    for j in range(q):
        working.kappa[j] += working.delta_kappa[j]
    working.state.lambda_ += delta_lambda
    working.state.mu += delta_mu
    working.state.upsilon += delta_upsilon

cpdef tuple knot_search(KnotSearchData data, FLOAT_t[:] candidates, FLOAT_t[:] p, INDEX_t q, INDEX_t m,
                 INDEX_t r, INDEX_t n_outcomes, int verbose):
    cdef KnotSearchReadOnlyData constant = data.constant
    cdef PredictorDependentData predictor = constant.predictor
    cdef list outcomes = constant.outcome.outcomes
    cdef list workings = data.workings

    # These assertions should be satisfied
    # assert len(outcomes) == n_outcomes
    # assert len(workings) == len(outcomes)
    # assert len(candidates) == r
    # assert outcomes[0].k == q

    # Initialize variables to their pre-loop values.  These are the values
    # they would have for a hypothetical knot candidate, phi, such that
    # phi > max(x).  This only matters for values that will be tracked and
    # updated across iterations.  Values that are calculated from scratch at
    # each iteration are not initialized.
    cdef FLOAT_t best_knot = 0.
    cdef INDEX_t best_knot_index = 0
    cdef FLOAT_t phi_next = candidates[0]
    cdef FLOAT_t phi
    cdef KnotSearchWorkingData working
    cdef INDEX_t j, i
    for j in range(n_outcomes):
        working = workings[j]
        working.state.phi_next = phi_next
        working.state.alpha = 0.
        working.state.beta = 0.
        for i in range(q):
            working.gamma[i] = 0.
        for i in range(q):
            working.kappa[i] = 0.
        working.state.lambda_ = 0.
        working.state.mu = 0.
        working.state.upsilon = 0.
        working.state.ord_idx = 0
        working.state.idx = predictor.order[working.state.ord_idx]

    # A lower bound for zeta_squared is 0 (it is the square of a real number),
    # so initialize best_zeta_squared to 0.
    best_zeta_squared = 0.

    # Iterate over candidates.
    # Loop invariant: At the start (and end) of each iteration, alpha, beta,
    # and gamma are correct for the knot phi_next.  Kappa, lambda, mu, and
    # upsilon are correct for the update from (not to) phi_next.  That is,
    # alpha, beta, and gamma should be updated before kappa, lambda, mu,
    # and upsilon are updated.
    cdef SingleOutcomeDependentData outcome
    cdef FLOAT_t zeta_squared
    cdef INDEX_t k
    cdef FLOAT_t loss
    for k in range(r):
        phi = phi_next
        phi_next = candidates[k]
        zeta_squared = 0.
        for i in range(n_outcomes):
            working = workings[i]
            outcome = outcomes[i]
#             omega_minus_theta_squared = outcome.omega - np.dot(outcome.theta[:q], outcome.theta[:q])

            # Get the next candidate knot
            working.state.phi = phi
            working.state.phi_next = phi_next

            # Update workingdata for the new candidate knot
            fast_update(predictor, outcome, working, p, q, m, r)

            if working.state.zeta_squared >= outcome.sse_:
                # Sometimes this can happen because of numerical issues in
                # the fast update process.  These occur when the new potential
                # predictor column is close to linear dependence on previous
                # columns.  In that case, correct everything so that we can move
                # on.
#                 print 'this is a problem!'
#                 print 'zeta_squared =', working.state.zeta_squared
#                 print 'omega_minus_theta_squared =', outcome.sse_
#                 print i
#                 print 'epsilon_squared =',  working.state.beta - np.dot(working.gamma[:q], working.gamma[:q])
#                 print 'alpha =', working.state.alpha
#                 print 'gamma * theta =', dot(working.gamma, outcome.theta, q)
#                 print 'beta =', working.state.beta
#                 print 'gamma^2 = ', dot(working.gamma, working.gamma, q)
                if verbose >= 2:
                    print('Encountered numerical problem in knot search.  The problem is being corrected by a slower computation.')
                    if verbose >= 3:
                        print('Potentially helpful numbers if you are really interested:')
                        print('zeta_squared = %f') % zeta_squared
                        print('omega_minus_theta_squared = %f' % outcome.sse_)
                        print('epsilon_squared =',  working.state.beta - np.dot(working.gamma[:q], working.gamma[:q]))
                        print('alpha =', working.state.alpha)
                        print('gamma * theta =', dot(working.gamma, outcome.theta, q))
                        print('beta =', working.state.beta)
                        print('gamma^2 = ', dot(working.gamma, working.gamma, q))
                        print('phi = ' % working.state.phi)
                        print('phi_next = ' % working.state.phi_next)
                        print('p = %d' % p)
                        print('q = %d' % q)
                        print('m = %d' % m)
                        print('r = %d' % r)
                        print('k = %d' % k)
                        print('i = %d' % i)
                working.state.zeta_squared = 0.
                working.state.alpha = dot(working.gamma, outcome.theta, q)
                working.state.beta = dot(working.gamma, working.gamma, q)


            # Add up objectives
            zeta_squared += working.state.zeta_squared
        # Compare against best result so far
        if zeta_squared > best_zeta_squared:
            best_knot_index = k
            best_knot = phi_next
            best_zeta_squared = zeta_squared

        # DEBUG
#         loss = -best_zeta_squared
#         for i in range(n_outcomes):
#             outcome = outcomes[i]
#             loss += outcome.omega - np.dot(outcome.theta[:q], outcome.theta[:q])
#         if loss < 0:
#             print 'negative loss!'
#             print 'best_zeta_squared =', best_zeta_squared
#             x_should_be = np.maximum(np.asarray(predictor.x) - best_knot, 0) * p
#
#             for i in range(n_outcomes):
#                 outcome = outcomes[i]
#                 print 'should be eye = ', np.dot(outcome.weight.Q_t[:q,:], np.transpose(outcome.weight.Q_t[:q,:]))
#                 omega_should_be = np.dot(np.array(outcome.weight.w) * outcome.y, np.array(outcome.weight.w) * outcome.y)
#                 theta_should_be = np.dot(outcome.weight.Q_t[:q,:], np.array(outcome.weight.w) * outcome.y)
#                 print 'omega =', outcome.omega, omega_should_be
#                 print 'theta =', np.array(outcome.theta[:q]), theta_should_be
#                 print 'theta^2 =', np.dot(outcome.theta[:q], outcome.theta[:q]), np.dot(theta_should_be, theta_should_be)
#             raise ValueError
        # END DEBUG
    # Calculate value of overall objective function
    # (this is the sqrt of the sum of squared residuals)
    loss = -best_zeta_squared
    for i in range(n_outcomes):
        outcome = outcomes[i]
        loss += outcome.sse_
#     if loss < 0:
#         print 'negative loss!'
#         print 'best_zeta_squared =', best_zeta_squared
#         x_should_be = np.maximum(np.asarray(predictor.x) - best_knot, 0) * p
#
#         for i in range(n_outcomes):
#             outcome = outcomes[i]
#
#             omega_should_be = np.dot(np.array(outcome.weight.w) * outcome.y, np.array(outcome.weight.w) * outcome.y)
#             theta_should_be = np.dot(outcome.weight.Q_t[:q,:], np.array(outcome.weight.w) * outcome.y)
#             print 'omega =', outcome.omega, omega_should_be
#             print 'theta =', np.array(outcome.theta[:q]), theta_should_be
#             print 'theta^2 =', np.dot(outcome.theta[:q], outcome.theta[:q]), np.dot(theta_should_be, theta_should_be)
#         raise ValueError
#     loss = sqrt(loss)

    # Return
    return best_knot, best_knot_index, loss



from __future__ import annotations

import pickle

import numpy

from pyearth._basis import ConstantBasisFunction, LinearBasisFunction
from pyearth._types import BOOL

from . import assert_equal, assert_true
from .base import BaseContainer


class Container(BaseContainer):
    def __init__(self):
        super(Container, self).__init__()
        self.parent = ConstantBasisFunction()
        self.bf = LinearBasisFunction(self.parent, 1)


def test_apply():
    cnt = Container()
    m, n = cnt.X.shape
    missing = numpy.zeros_like(cnt.X, dtype=BOOL)
    B = numpy.ones(shape=(m, 10))
    cnt.bf.apply(cnt.X, missing, B[:, 0])
    assert_true(numpy.all(B[:, 0] == cnt.X[:, 1]))


def test_apply_deriv():
    cnt = Container()
    m, _ = cnt.X.shape
    missing = numpy.zeros_like(cnt.X, dtype=BOOL)
    b = numpy.empty(shape=m)
    j = numpy.empty(shape=m)
    cnt.bf.apply_deriv(cnt.X, missing, b, j, 1)
    numpy.testing.assert_almost_equal(b, cnt.X[:, 1])
    numpy.testing.assert_almost_equal(j, 1.0)


def test_degree():
    cnt = Container()
    assert_equal(cnt.bf.degree(), 1)


def test_pickle_compatibility():
    cnt = Container()
    bf_copy = pickle.loads(pickle.dumps(cnt.bf))
    assert_true(cnt.bf == bf_copy)


def test_smoothed_version():
    cnt = Container()
    translation = {cnt.parent: cnt.parent._smoothed_version(None, {}, {})}
    smoothed = cnt.bf._smoothed_version(cnt.parent, {}, translation)
    assert_true(isinstance(smoothed, LinearBasisFunction))
    assert_equal(smoothed.get_variable(), cnt.bf.get_variable())



"""
================================================================
Plotting sine function with redundant predictors an missing data
================================================================

An example plotting a fit of the sine function.  There are two
redundant predictors, each of which has independent and random
missingness.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy

from pyearth import Earth

# Create some fake data
numpy.random.seed(2)
m = 10000
n = 10
X = 80 * numpy.random.uniform(size=(m, n)) - 40
X[:, 5] = X[:, 6] + numpy.random.normal(0, 0.1, m)
y = 100 * (numpy.sin((X[:, 5] + X[:, 6]) / 20) - 4.0) + 10 * numpy.random.normal(size=m)
missing = numpy.random.binomial(1, 0.2, (m, n)).astype(bool)
X_full = X.copy()
X[missing] = None
idx5 = (1 - missing[:, 5]).astype(bool)
idx6 = (1 - missing[:, 6]).astype(bool)

# Fit an Earth model
model = Earth(
    max_degree=5,
    minspan_alpha=0.5,
    allow_missing=True,
    enable_pruning=True,
    thresh=0.001,
    smooth=True,
    verbose=2,
)
model.fit(X, y)
# Print the model
print(model.summary())

# Plot the model
y_hat = model.predict(X)
fig = plt.figure()

ax1 = fig.add_subplot(3, 2, 1)
ax1.plot(X_full[idx5, 5], y[idx5], "b.")
ax1.plot(X_full[idx5, 5], y_hat[idx5], "r.")
ax1.set_xlim(-40, 40)
ax1.set_title("x5 present")
ax1.set_xlabel("x5")

ax2 = fig.add_subplot(3, 2, 2)
ax2.plot(X_full[idx6, 6], y[idx6], "b.")
ax2.plot(X_full[idx6, 6], y_hat[idx6], "r.")
ax2.set_xlim(-40, 40)
ax2.set_title("x6 present")
ax2.set_xlabel("x6")

ax3 = fig.add_subplot(3, 2, 3, sharex=ax1)
ax3.plot(X_full[~idx6, 5], y[~idx6], "b.")
ax3.plot(X_full[~idx6, 5], y_hat[~idx6], "r.")
ax3.set_title("x6 missing")
ax3.set_xlabel("x5")

ax4 = fig.add_subplot(3, 2, 4, sharex=ax2)
ax4.plot(X_full[~idx5, 6], y[~idx5], "b.")
ax4.plot(X_full[~idx5, 6], y_hat[~idx5], "r.")
ax4.set_title("x5 missing")
ax4.set_xlabel("x6")

ax5 = fig.add_subplot(3, 2, 5, sharex=ax1)
ax5.plot(X_full[(~idx6) & (~idx5), 5], y[(~idx6) & (~idx5)], "b.")
ax5.plot(X_full[(~idx6) & (~idx5), 5], y_hat[(~idx6) & (~idx5)], "r.")
ax5.set_title("both missing")
ax5.set_xlabel("x5")

ax6 = fig.add_subplot(3, 2, 6, sharex=ax2)
ax6.plot(X_full[(~idx6) & (~idx5), 6], y[(~idx6) & (~idx5)], "b.")
ax6.plot(X_full[(~idx6) & (~idx5), 6], y_hat[(~idx6) & (~idx5)], "r.")
ax6.set_title("both missing")
ax6.set_xlabel("x6")

fig.tight_layout()
plt.show()



"""
======================================================
Plotting sckit-learn classifiers comparison with Earth
======================================================

This script recreates the scikit-learn classifier comparison example found at
http://scikit-learn.org/stable/auto_examples/classification/plot_classifier_comparison.html.
It has been modified to include an Earth based classifier.
"""

# Code source: Gael Varoqueux
#              Andreas Mueller
# Modified for Documentation merge by Jacques Grobler
# License: BSD 3 clause
# Modified to include pyearth by Jason Rudy
from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import ListedColormap
from sklearn.cross_validation import train_test_split
from sklearn.datasets import make_circles, make_classification, make_moons
from sklearn.ensemble import RandomForestClassifier
from sklearn.lda import LDA
from sklearn.linear_model.logistic import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.qda import QDA
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier

from pyearth.earth import Earth

print(__doc__)

h = 0.02  # step size in the mesh

np.random.seed(1)

# Combine Earth with LogisticRegression in a pipeline to do classification
earth_classifier = Pipeline([("earth", Earth(max_degree=3, penalty=1.5)), ("logistic", LogisticRegression())])

names = [
    "Nearest Neighbors",
    "Linear SVM",
    "RBF SVM",
    "Decision Tree",
    "Random Forest",
    "Naive Bayes",
    "LDA",
    "QDA",
    "Earth",
]
classifiers = [
    KNeighborsClassifier(3),
    SVC(kernel="linear", C=0.025, probability=True),
    SVC(gamma=2, C=1, probability=True),
    DecisionTreeClassifier(max_depth=5),
    RandomForestClassifier(max_depth=5, n_estimators=10, max_features=1),
    GaussianNB(),
    LDA(),
    QDA(),
    earth_classifier,
]

X, y = make_classification(n_features=2, n_redundant=0, n_informative=2, random_state=1, n_clusters_per_class=1)
rng = np.random.RandomState(2)
X += 2 * rng.uniform(size=X.shape)
linearly_separable = (X, y)

datasets = [
    make_moons(noise=0.3, random_state=0),
    make_circles(noise=0.2, factor=0.5, random_state=1),
    linearly_separable,
]

figure = plt.figure(figsize=(27, 9))
i = 1
# iterate over datasets
for ds in datasets:
    # preprocess dataset, split into training and test part
    X, y = ds
    X = StandardScaler().fit_transform(X)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.4)

    x_min, x_max = X[:, 0].min() - 0.5, X[:, 0].max() + 0.5
    y_min, y_max = X[:, 1].min() - 0.5, X[:, 1].max() + 0.5
    xx, yy = np.meshgrid(np.arange(x_min, x_max, h), np.arange(y_min, y_max, h))

    # just plot the dataset first
    cm = plt.cm.RdBu
    cm_bright = ListedColormap(["#FF0000", "#0000FF"])
    ax = plt.subplot(len(datasets), len(classifiers) + 1, i)
    # Plot the training points
    ax.scatter(X_train[:, 0], X_train[:, 1], c=y_train, cmap=cm_bright)
    # and testing points
    ax.scatter(X_test[:, 0], X_test[:, 1], c=y_test, cmap=cm_bright, alpha=0.6)
    ax.set_xlim(xx.min(), xx.max())
    ax.set_ylim(yy.min(), yy.max())
    ax.set_xticks(())
    ax.set_yticks(())
    i += 1

    # iterate over classifiers
    for name, clf in zip(names, classifiers, strict=False):
        ax = plt.subplot(len(datasets), len(classifiers) + 1, i)
        clf.fit(X_train, y_train)
        score = clf.score(X_test, y_test)

        # Plot the decision boundary. For that, we will assign a color to each
        # point in the mesh [x_min, m_max]x[y_min, y_max].
        try:
            Z = clf.predict_proba(np.c_[xx.ravel(), yy.ravel()])[:, 1]
        except NotImplementedError:
            Z = clf.decision_function(np.c_[xx.ravel(), yy.ravel()])

        # Put the result into a color plot
        Z = Z.reshape(xx.shape)
        ax.contourf(xx, yy, Z, cmap=cm, alpha=0.8)

        # Plot also the training points
        ax.scatter(X_train[:, 0], X_train[:, 1], c=y_train, cmap=cm_bright)
        # and testing points
        ax.scatter(X_test[:, 0], X_test[:, 1], c=y_test, cmap=cm_bright, alpha=0.6)

        ax.set_xlim(xx.min(), xx.max())
        ax.set_ylim(yy.min(), yy.max())
        ax.set_xticks(())
        ax.set_yticks(())
        ax.set_title(name)
        ax.text(
            xx.max() - 0.3,
            yy.min() + 0.3,
            ("%.2f" % score).lstrip("0"),
            size=15,
            horizontalalignment="right",
        )
        i += 1

figure.subplots_adjust(left=0.02, right=0.98)
plt.savefig("classifier_comp.pdf", transparent=True)
plt.show()



"""
======================================================
Plotting a linear function with a categorical variable
======================================================

Fitting a pyearth model to a linear function shows that pyearth
will automatically  choose a linear basis function in some cases.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np

from pyearth import Earth

np.random.seed(1)
m = 1000
n = 5

X = np.random.normal(size=(m, n))

# Make X[:,1] binary
X[:, 1] = np.random.binomial(1, 0.5, size=m)

# The response is a linear function of the inputs
y = 2 * X[:, 0] + 3 * X[:, 1] + np.random.normal(size=m)

# Fit the earth model
model = Earth().fit(X, y)

# Print the model summary, showing linear terms
print(model.summary())

# Plot for both values of X[:,1]
y_hat = model.predict(X)
plt.figure()
plt.plot(X[:, 0], y, "k.")
plt.plot(X[X[:, 1] == 0, 0], y_hat[X[:, 1] == 0], "r.", label="$x_1 = 0$")
plt.plot(X[X[:, 1] == 1, 0], y_hat[X[:, 1] == 1], "b.", label="$x_1 = 1$")
plt.legend(loc="best")
plt.xlabel("$x_0$")
plt.show()



"""
=================================================================
Demonstrating a use of weights in outputs with two sine functions
=================================================================

Each row in the grid is a run of an earth model.
Each column is an output.
In each run, different weights are given to
the outputs.

"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np

from pyearth import Earth

# Create some fake data
np.random.seed(2)
m = 10000
n = 10
X = 80 * np.random.uniform(size=(m, n)) - 40
y1 = 120 * np.abs(np.sin((X[:, 6]) / 6) - 1.0) + 15 * np.random.normal(size=m)
y2 = 120 * np.abs(np.sin((X[:, 5]) / 6) - 1.0) + 15 * np.random.normal(size=m)

y1 = (y1 - y1.mean()) / y1.std()
y2 = (y2 - y2.mean()) / y2.std()
y_mix = np.concatenate((y1[:, np.newaxis], y2[:, np.newaxis]), axis=1)

alphas = [0.9, 0.8, 0.6, 0.4, 0.2, 0.1]
n_plots = len(alphas)
k = 1
fig = plt.figure(figsize=(10, 15))
for i, alpha in enumerate(alphas):
    # Fit an Earth model
    model = Earth(
        max_degree=5,
        minspan_alpha=0.05,
        endspan_alpha=0.05,
        max_terms=10,
        check_every=1,
        thresh=0.0,
    )
    output_weight = np.array([alpha, 1 - alpha])
    model.fit(X, y_mix, output_weight=output_weight)
    print(model.summary())

    # Plot the model
    y_hat = model.predict(X)

    mse = ((y_hat - y_mix) ** 2).mean(axis=0)
    ax = plt.subplot(n_plots, 2, k)
    ax.set_ylabel("Run {0}".format(i + 1), rotation=0, labelpad=20)
    plt.plot(X[:, 6], y_mix[:, 0], "r.")
    plt.plot(X[:, 6], model.predict(X)[:, 0], "b.")
    plt.title("MSE: {0:.3f}, Weight : {1:.1f}".format(mse[0], alpha))
    plt.subplot(n_plots, 2, k + 1)
    plt.plot(X[:, 5], y_mix[:, 1], "r.")
    plt.plot(X[:, 5], model.predict(X)[:, 1], "b.")
    plt.title("MSE: {0:.3f}, Weight : {1:.1f}".format(mse[1], 1 - alpha))
    k += 2
plt.tight_layout()
plt.show()



"""
==================================
Plotting two simple sine functions
==================================

A simple example plotting a fit of two sine functions.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy

from pyearth import Earth

# Create some fake data
numpy.random.seed(2)
m = 10000
n = 10
X = 80 * numpy.random.uniform(size=(m, n)) - 40
y1 = 100 * numpy.abs(numpy.sin((X[:, 6]) / 10) - 4.0) + 10 * numpy.random.normal(size=m)

y2 = 100 * numpy.abs(numpy.sin((X[:, 6]) / 2) - 8.0) + 5 * numpy.random.normal(size=m)

# Fit an Earth model
model = Earth(max_degree=3, minspan_alpha=0.5)
y_mix = numpy.concatenate((y1[:, numpy.newaxis], y2[:, numpy.newaxis]), axis=1)
model.fit(X, y_mix)

# Print the model
print(model.trace())
print(model.summary())

# Plot the model
y_hat = model.predict(X)

fig = plt.figure()

ax = fig.add_subplot(1, 2, 1)
ax.plot(X[:, 6], y_mix[:, 0], "r.")
ax.plot(X[:, 6], model.predict(X)[:, 0], "b.")

ax = fig.add_subplot(1, 2, 2)
ax.plot(X[:, 6], y_mix[:, 1], "r.")
ax.plot(X[:, 6], model.predict(X)[:, 1], "b.")
plt.show()



"""
XKCD plot generator
-------------------
Author: Jake Vanderplas

This is a script that will take any matplotlib line diagram, and convert it
to an XKCD-style plot.  It will work for plots with line & text elements,
including axes labels and titles (but not axes tick labels).

The idea for this comes from work by Damon McDougall
  http://www.mail-archive.com/matplotlib-users@lists.sourceforge.net/msg25499.html

Copied from Jake's blog:
http://jakevdp.github.com/blog/2012/10/07/xkcd-style-plots-in-matplotlib/
on 2013-03-07

"""

# We need a special font for the code below.  It can be downloaded this way:
from __future__ import annotations

import os
import urllib

import matplotlib.font_manager as fm
import numpy as np
import pylab as pl
from scipy import interpolate, signal

if not os.path.exists("Humor-Sans.ttf"):
    fhandle = urllib.urlopen("https://github.com/shreyankg/xkcd-desktop/raw/master/Humor-Sans.ttf")
    open("Humor-Sans.ttf", "wb").write(fhandle.read())


def xkcd_line(x, y, xlim=None, ylim=None, mag=1.0, f1=30, f2=0.05, f3=15):
    """
    Mimic a hand-drawn line from (x, y) data

    Parameters
    ----------
    x, y : array_like
        arrays to be modified
    xlim, ylim : data range
        the assumed plot range for the modification.  If not specified,
        they will be guessed from the  data
    mag : float
        magnitude of distortions
    f1, f2, f3 : int, float, int
        filtering parameters.  f1 gives the size of the window, f2 gives
        the high-frequency cutoff, f3 gives the size of the filter

    Returns
    -------
    x, y : ndarrays
        The modified lines
    """
    x = np.asarray(x)
    y = np.asarray(y)

    # get limits for rescaling
    if xlim is None:
        xlim = (x.min(), x.max())
    if ylim is None:
        ylim = (y.min(), y.max())

    if xlim[1] == xlim[0]:
        xlim = ylim

    if ylim[1] == ylim[0]:
        ylim = xlim

    # scale the data
    x_scaled = (x - xlim[0]) * 1.0 / (xlim[1] - xlim[0])
    y_scaled = (y - ylim[0]) * 1.0 / (ylim[1] - ylim[0])

    # compute the total distance along the path
    dx = x_scaled[1:] - x_scaled[:-1]
    dy = y_scaled[1:] - y_scaled[:-1]
    dist_tot = np.sum(np.sqrt(dx * dx + dy * dy))

    # number of interpolated points is proportional to the distance
    Nu = int(200 * dist_tot)
    u = np.arange(-1, Nu + 1) * 1.0 / (Nu - 1)

    # interpolate curve at sampled points
    k = min(3, len(x) - 1)
    res = interpolate.splprep([x_scaled, y_scaled], s=0, k=k)
    x_int, y_int = interpolate.splev(u, res[0])

    # we'll perturb perpendicular to the drawn line
    dx = x_int[2:] - x_int[:-2]
    dy = y_int[2:] - y_int[:-2]
    dist = np.sqrt(dx * dx + dy * dy)

    # create a filtered perturbation
    coeffs = mag * np.random.normal(0, 0.01, len(x_int) - 2)
    b = signal.firwin(f1, f2 * dist_tot, window=("kaiser", f3))
    response = signal.lfilter(b, 1, coeffs)

    x_int[1:-1] += response * dy / dist
    y_int[1:-1] += response * dx / dist

    # un-scale data
    x_int = x_int[1:-1] * (xlim[1] - xlim[0]) + xlim[0]
    y_int = y_int[1:-1] * (ylim[1] - ylim[0]) + ylim[0]

    return x_int, y_int


def XKCDify(
    ax,
    mag=1.0,
    f1=50,
    f2=0.01,
    f3=15,
    bgcolor="w",
    xaxis_loc=None,
    yaxis_loc=None,
    xaxis_arrow="+",
    yaxis_arrow="+",
    ax_extend=0.1,
    expand_axes=False,
):
    """Make axis look hand-drawn

    This adjusts all lines, text, legends, and axes in the figure to look
    like xkcd plots.  Other plot elements are not modified.

    Parameters
    ----------
    ax : Axes instance
        the axes to be modified.
    mag : float
        the magnitude of the distortion
    f1, f2, f3 : int, float, int
        filtering parameters.  f1 gives the size of the window, f2 gives
        the high-frequency cutoff, f3 gives the size of the filter
    xaxis_loc, yaxis_log : float
        The locations to draw the x and y axes.  If not specified, they
        will be drawn from the bottom left of the plot
    xaxis_arrow, yaxis_arrow : str
        where to draw arrows on the x/y axes.  Options are '+', '-', '+-', or ''
    ax_extend : float
        How far (fractionally) to extend the drawn axes beyond the original
        axes limits
    expand_axes : bool
        if True, then expand axes to fill the figure (useful if there is only
        a single axes in the figure)
    """
    # Get axes aspect
    ext = ax.get_window_extent().extents
    aspect = (ext[3] - ext[1]) / (ext[2] - ext[0])

    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    xspan = xlim[1] - xlim[0]
    yspan = ylim[1] - xlim[0]

    xax_lim = (xlim[0] - ax_extend * xspan, xlim[1] + ax_extend * xspan)
    yax_lim = (ylim[0] - ax_extend * yspan, ylim[1] + ax_extend * yspan)

    if xaxis_loc is None:
        xaxis_loc = ylim[0]

    if yaxis_loc is None:
        yaxis_loc = xlim[0]

    # Draw axes
    xaxis = pl.Line2D([xax_lim[0], xax_lim[1]], [xaxis_loc, xaxis_loc], linestyle="-", color="k")
    yaxis = pl.Line2D([yaxis_loc, yaxis_loc], [yax_lim[0], yax_lim[1]], linestyle="-", color="k")

    # Label axes3, 0.5, 'hello', fontsize=14)
    ax.text(
        xax_lim[1],
        xaxis_loc - 0.02 * yspan,
        ax.get_xlabel(),
        fontsize=14,
        ha="right",
        va="top",
        rotation=12,
    )
    ax.text(
        yaxis_loc - 0.02 * xspan,
        yax_lim[1],
        ax.get_ylabel(),
        fontsize=14,
        ha="right",
        va="top",
        rotation=78,
    )
    ax.set_xlabel("")
    ax.set_ylabel("")

    # Add title
    ax.text(
        0.5 * (xax_lim[1] + xax_lim[0]),
        yax_lim[1],
        ax.get_title(),
        ha="center",
        va="bottom",
        fontsize=16,
    )
    ax.set_title("")

    Nlines = len(ax.lines)
    lines = [xaxis, yaxis] + [ax.lines.pop(0) for i in range(Nlines)]

    for line in lines:
        x, y = line.get_data()

        x_int, y_int = xkcd_line(x, y, xlim, ylim, mag, f1, f2, f3)

        # create foreground and background line
        lw = line.get_linewidth()
        line.set_linewidth(2 * lw)
        line.set_data(x_int, y_int)

        # don't add background line for axes
        if (line is not xaxis) and (line is not yaxis):
            line_bg = pl.Line2D(x_int, y_int, color=bgcolor, linewidth=8 * lw)

            ax.add_line(line_bg)
        ax.add_line(line)

    # Draw arrow-heads at the end of axes lines
    arr1 = 0.03 * np.array([-1, 0, -1])
    arr2 = 0.02 * np.array([-1, 0, 1])

    arr1[::2] += np.random.normal(0, 0.005, 2)
    arr2[::2] += np.random.normal(0, 0.005, 2)

    x, y = xaxis.get_data()
    if "+" in str(xaxis_arrow):
        ax.plot(x[-1] + arr1 * xspan * aspect, y[-1] + arr2 * yspan, color="k", lw=2)
    if "-" in str(xaxis_arrow):
        ax.plot(x[0] - arr1 * xspan * aspect, y[0] - arr2 * yspan, color="k", lw=2)

    x, y = yaxis.get_data()
    if "+" in str(yaxis_arrow):
        ax.plot(x[-1] + arr2 * xspan * aspect, y[-1] + arr1 * yspan, color="k", lw=2)
    if "-" in str(yaxis_arrow):
        ax.plot(x[0] - arr2 * xspan * aspect, y[0] - arr1 * yspan, color="k", lw=2)

    # Change all the fonts to humor-sans.
    prop = fm.FontProperties(fname="Humor-Sans.ttf", size=16)
    for text in ax.texts:
        text.set_fontproperties(prop)

    # modify legend
    leg = ax.get_legend()
    if leg is not None:
        leg.set_frame_on(False)

        for child in leg.get_children():
            if isinstance(child, pl.Line2D):
                x, y = child.get_data()
                child.set_data(xkcd_line(x, y, mag=10, f1=100, f2=0.001))
                child.set_linewidth(2 * child.get_linewidth())
            if isinstance(child, pl.Text):
                child.set_fontproperties(prop)

    # Set the axis limits
    ax.set_xlim(xax_lim[0] - 0.1 * xspan, xax_lim[1] + 0.1 * xspan)
    ax.set_ylim(yax_lim[0] - 0.1 * yspan, yax_lim[1] + 0.1 * yspan)

    # adjust the axes
    ax.set_xticks([])
    ax.set_yticks([])

    if expand_axes:
        ax.figure.set_facecolor(bgcolor)
        ax.set_axis_off()
        ax.set_position([0, 0, 1, 1])

    return ax



from __future__ import annotations

import pickle

import numpy

from pyearth._basis import (
    ConstantBasisFunction,
    HingeBasisFunction,
    SmoothedHingeBasisFunction,
)
from pyearth._types import BOOL

from . import assert_equal, assert_true
from .base import BaseContainer


class Container(BaseContainer):
    def __init__(self):
        super(Container, self).__init__()
        self.parent = ConstantBasisFunction()
        self.bf = HingeBasisFunction(self.parent, 1.0, 10, 1, False)


def test_getters():
    cnt = Container()
    assert not cnt.bf.get_reverse()
    assert cnt.bf.get_knot() == 1.0
    assert cnt.bf.get_variable() == 1
    assert cnt.bf.get_knot_idx() == 10
    assert cnt.bf.get_parent() == cnt.parent


def test_apply():
    cnt = Container()
    m, _ = cnt.X.shape
    missing = numpy.zeros_like(cnt.X, dtype=BOOL)
    B = numpy.ones(shape=(m, 10))
    cnt.bf.apply(cnt.X, missing, B[:, 0])
    numpy.testing.assert_almost_equal(B[:, 0], (cnt.X[:, 1] - 1.0) * (cnt.X[:, 1] > 1.0))


def test_apply_deriv():
    cnt = Container()
    m, _ = cnt.X.shape
    missing = numpy.zeros_like(cnt.X, dtype=BOOL)
    b = numpy.empty(shape=m)
    j = numpy.empty(shape=m)
    cnt.bf.apply_deriv(cnt.X, missing, b, j, 1)
    numpy.testing.assert_almost_equal((cnt.X[:, 1] - 1.0) * (cnt.X[:, 1] > 1.0), b)
    numpy.testing.assert_almost_equal(1.0 * (cnt.X[:, 1] > 1.0), j)


def test_degree():
    cnt = Container()
    assert_equal(cnt.bf.degree(), 1)


def test_pickle_compatibility():
    cnt = Container()
    bf_copy = pickle.loads(pickle.dumps(cnt.bf))
    assert_true(cnt.bf == bf_copy)


def test_smoothed_version():
    cnt = Container()
    knot_dict = {cnt.bf: (0.5, 1.5)}
    translation = {cnt.parent: cnt.parent._smoothed_version(None, {}, {})}
    smoothed = cnt.bf._smoothed_version(cnt.parent, knot_dict, translation)

    assert_true(type(smoothed) is SmoothedHingeBasisFunction)
    assert_true(translation[cnt.parent] is smoothed.get_parent())
    assert_equal(smoothed.get_knot_minus(), 0.5)
    assert_equal(smoothed.get_knot_plus(), 1.5)
    assert_equal(smoothed.get_knot(), cnt.bf.get_knot())
    assert_equal(smoothed.get_variable(), cnt.bf.get_variable())



from __future__ import annotations

import matplotlib as mpl

mpl.use("Agg")
import numpy
from matplotlib import pyplot
from xkcdify import XKCDify

from pyearth import Earth

# =========================================================================
# V-Function Example
# =========================================================================
# Create some fake data
numpy.random.seed(0)
m = 1000
n = 10
X = 80 * numpy.random.uniform(size=(m, n)) - 40
y = numpy.abs(X[:, 6] - 4.0) + 1 * numpy.random.normal(size=m)

# Fit an Earth model
model = Earth()
model.fit(X, y)

# Print the model
print(model.trace())
print(model.summary())

# Plot the model
y_hat = model.predict(X)
pyplot.figure()
pyplot.plot(X[:, 6], y, "r.")
pyplot.plot(X[:, 6], y_hat, "b.")
pyplot.xlabel("x_6")
pyplot.ylabel("y")
pyplot.title("Simple Earth Example")
pyplot.savefig("simple_earth_example.png")

# =========================================================================
# Hinge plot
# =========================================================================

x = numpy.arange(-10, 10, 0.1)
y = x * (x > 0)

fig = pyplot.figure(figsize=(10, 5))
pyplot.plot(x, y)
ax = pyplot.gca()

pyplot.title("Basic Hinge Function")
pyplot.xlabel("x")
pyplot.ylabel("h(x)")
pyplot.annotate(
    "x=t",
    (0, 0),
    xytext=(-30, 30),
    textcoords="offset points",
    arrowprops=dict(arrowstyle="->", connectionstyle="arc3,rad=.2"),
)
XKCDify(ax)
pyplot.setp(ax, frame_on=False)
pyplot.savefig("hinge.png")

# =========================================================================
# Piecewise Linear Plot
# =========================================================================
m = 1000
x = numpy.arange(-10, 10, 0.1)
y = 1 - 2 * (1 - x) * (x < 1) + 0.5 * (x - 1) * (x > 1)

pyplot.figure(figsize=(10, 5))
pyplot.plot(x, y)
ax = pyplot.gca()
pyplot.xlabel("x")
pyplot.ylabel("y")
pyplot.title("Piecewise Linear Function")
XKCDify(ax)
pyplot.setp(ax, frame_on=False)
pyplot.savefig("piecewise_linear.png")



from __future__ import annotations


def export_python_function(earth_model):
    """
    Exports model as a pure python function, with no numpy/scipy/sklearn dependencies.
    :param earth_model: Trained pyearth model
    :return: A function that accepts an iterator over examples, and returns an iterator over transformed examples
    """
    i = 0
    accessors = []
    for bf in earth_model.basis_:
        if not bf.is_pruned():
            accessors.append(bf.func_factory(earth_model.coef_[0, i]))
            i += 1

    def func(example_iterator):
        return [sum(accessor(row) for accessor in accessors) for row in example_iterator]

    return func


def export_python_string(earth_model, function_name="model"):
    """
    Exports model as a string that evaluates as python code, with no numpy/scipy/sklearn dependencies.
    :param earth_model: Trained pyearth model
    :param function_name: string, optional, will be the name of the function in the returned string
    :return: string, when executed (either by writing to a file, or using `exec`, will define a python
      function that accepts an iterator over examples, and returns an iterator over transformed examples
    """
    i = 0
    accessors = []
    for bf in earth_model.basis_:
        if not bf.is_pruned():
            accessors.append(bf.func_string_factory(earth_model.coef_[0, i]))
            i += 1

    return """def {:s}(example_iterator):
    accessors = [{:s}]
    for x in example_iterator:
        yield sum(accessor(x) for accessor in accessors)
    """.format(function_name, ",\n\t\t".join(accessors))


def export_sympy_term_expressions(earth_model):
    """
    Construct a list of sympy expressions for all non-pruned terms in the model.

    :param earth_model: Trained pyearth model
    :return: a list of sympy expressions representing terms in the model.  These
      expressions are the symbolic equivalent of the Earth.transform method.

    """
    from sympy import (
        Add,
        And,
        Function,
        Max,
        Mul,
        Not,
        Piecewise,
        Pow,
        RealNumber,
        Symbol,
    )

    from ._basis import (
        ConstantBasisFunction,
        HingeBasisFunction,
        LinearBasisFunction,
        MissingnessBasisFunction,
        SmoothedHingeBasisFunction,
        VariableBasisFunction,
    )

    Missing = Function("Missing")
    NaNProtect = Function("NaNProtect")

    def linear_bf_to_factor(bf, bf_var):
        return bf_var

    def smoothed_hinge_bf_to_factor(bf, bf_var):
        knot = RealNumber(bf.get_knot())
        knot_minus = RealNumber(bf.get_knot_minus())
        knot_plus = RealNumber(bf.get_knot_plus())
        r = RealNumber(bf.get_r())
        p = RealNumber(bf.get_p())
        if bf.get_reverse():
            lower_p = (-(bf_var - knot)), (bf_var <= knot_minus)
            upper_p = (0, bf_var >= knot_plus)
            left_exp = Mul(p, Pow((bf_var - knot_plus), 2))
            right_exp = Mul(r, Pow((bf_var - knot_plus), 3))
            middle_b = And(knot_minus < bf_var, bf_var < knot_plus)
            middle_exp = (Add(left_exp, right_exp), middle_b)
            piecewise = Piecewise(lower_p, upper_p, middle_exp)
            factor = piecewise
        else:
            lower_p = (0, bf_var <= knot_minus)
            upper_p = (bf_var - knot, bf_var >= knot_plus)
            left_exp = Mul(p, Pow((bf_var - knot_minus), 2))
            right_exp = Mul(r, Pow((bf_var - knot_minus), 3))
            middle_b = And(knot_minus < bf_var, bf_var < knot_plus)
            middle_exp = (Add(left_exp, right_exp), middle_b)
            piecewise = Piecewise(lower_p, upper_p, middle_exp)
            factor = piecewise
        return factor

    def hinge_bf_to_factor(bf, bf_var):
        knot = bf.get_knot()
        if bf.get_reverse():
            factor = Max(0, RealNumber(knot) - bf_var)
        else:
            factor = Max(0, bf_var - RealNumber(knot))
        return factor

    def missingness_bf_to_factor(bf, bf_var):
        # This is the error that should be raised when a user attempts to use functionality
        # that has not yet been implemented.
        if bf.complement:
            return Not(Missing(bf_var))
        else:
            return Missing(bf_var)

    def constant_bf_to_factor(bf, bf_var):
        return RealNumber(1)

    def protect_from_nan(label, missables):
        return NaNProtect(Symbol(label)) if label in missables else Symbol(label)

    def dont_protect_from_nan(label, missables):
        return Symbol(label)

    bf_to_factor_dispatcher = {
        LinearBasisFunction: linear_bf_to_factor,
        SmoothedHingeBasisFunction: smoothed_hinge_bf_to_factor,
        HingeBasisFunction: hinge_bf_to_factor,
        MissingnessBasisFunction: missingness_bf_to_factor,
        ConstantBasisFunction: constant_bf_to_factor,
    }

    nan_protect_dispatch = {
        LinearBasisFunction: protect_from_nan,
        SmoothedHingeBasisFunction: protect_from_nan,
        HingeBasisFunction: protect_from_nan,
        MissingnessBasisFunction: dont_protect_from_nan,
        ConstantBasisFunction: protect_from_nan,
    }

    def bf_to_factor(bf, missables):
        """
        Convert a BasisFunction to a factor of a term.
        """
        if isinstance(bf, VariableBasisFunction):
            bf_var = nan_protect_dispatch[bf.__class__](bf.label, missables)

        else:
            bf_var = None
        return bf_to_factor_dispatcher[bf.__class__](bf, bf_var)

    def missingness_bf_get_missables(bf):
        bf_var = bf.label
        return set([bf_var])

    def non_miscible(bf):
        return set()

    bf_get_missables_dispatcher = {
        LinearBasisFunction: non_miscible,
        SmoothedHingeBasisFunction: non_miscible,
        HingeBasisFunction: non_miscible,
        MissingnessBasisFunction: missingness_bf_get_missables,
        ConstantBasisFunction: non_miscible,
    }

    def get_missables(bf):
        missables = bf_get_missables_dispatcher[bf.__class__](bf)
        parent = bf.get_parent()
        if parent is None:
            return missables
        else:
            missables.update(get_missables(parent))

        return missables

    def bf_to_term(bf, missables):
        """
        Convert a BasisFunction to a term (without coefficient).
        """
        term = bf_to_factor(bf, missables)
        parent = bf.get_parent()
        if parent is None:
            return term
        else:
            return Mul(term, bf_to_term(parent, missables))

    return [bf_to_term(bf, get_missables(bf)) for bf in earth_model.basis_.piter()]


def export_sympy(earth_model, columns=None):
    """
    Constructs a sympy expression or list of sympy expressions from of a trained earth model.

    :param earth_model: Trained pyearth model
    :param columns: The index or indices of the output columns for which expressions are to
      be constructed.  If an integer is used, a sympy expression is returned.  If indices
      are given then a list of sympy expressions is returned.  If columns is None, it is treated
      as if columns=0 for models with only one output column or as columns=slice(None) for more than
      one output column.
    :return: a sympy expression or list of sympy expressions equivalent to the Earth.predict method for
      the selected output columns.

    """
    # Set a sane default for columns
    if columns is None:
        if earth_model.coef_.shape[0] == 1:
            columns = 0
        else:
            columns = slice(None)

    # Get basis function terms
    terms = export_sympy_term_expressions(earth_model)

    # Handle column choice
    coefs = earth_model.coef_[columns]
    if len(coefs.shape) == 1:
        unwrap = True
        coefs = [coefs]
        n_cols = 1
    else:
        unwrap = False
        n_cols = coefs.shape[0]

    # Combine coefficients with terms for each output column
    result = [sum([coefs[i][j] * term for j, term in enumerate(terms)]) for i in range(n_cols)]

    if unwrap:
        # Result should be an expression rather than a list of expressions.
        result = result[0]
    return result



# distutils: language = c
# cython: cdivision = True
# cython: boundscheck = False
# cython: wraparound = False
# cython: profile = False
# cython: language_level=3str
import numpy as np
from scipy.linalg.cython_lapack cimport dlarfg, dlarft, dlarfb
from scipy.linalg.cython_blas cimport dcopy
from libc.math cimport abs
from ._types import BOOL, FLOAT

cdef class UpdatingQT:
    def __init__(UpdatingQT self, int m, int max_n, Householder householder,
                 int k, FLOAT_t[::1, :] Q_t, FLOAT_t zero_tol, BOOL_t[::1] dependent_cols):
        self.m = m
        self.max_n = max_n
        self.householder = householder
        self.k = k
        self.Q_t = Q_t
        self.zero_tol = zero_tol
        self.dependent_cols = dependent_cols

    @classmethod
    def alloc(cls, int m, int max_n, FLOAT_t zero_tol):
        cdef Householder householder = Householder.alloc(m, max_n, zero_tol)
        cdef FLOAT_t[::1, :] Q_t = np.empty(shape=(max_n, m), dtype=FLOAT, order='F')
        cdef BOOL_t[::1] dependent_cols = np.empty(shape=max_n, dtype=BOOL, order='F')
        return cls(m, max_n, householder, 0, Q_t, zero_tol, dependent_cols)

    cpdef void update_qt(UpdatingQT self, bint dependent):
        # Assume that householder has already been updated and now Q_t needs to be updated
        # accordingly

        # Zero out the new row of Q_t
        cdef FLOAT_t zero = 0.
        cdef int zero_int = 0
        cdef int N = self.m
        cdef FLOAT_t * y = <FLOAT_t *> &(self.Q_t[self.k, 0])
        cdef int incy = self.max_n
        dcopy(&N, &zero, &zero_int, y, &incy)

        if not dependent:

            # Place a one in the right place
            # In general self.householder.k <= self.k + 1.
            # They are not necessarily equal.
            self.Q_t[self.k, self.householder.k - 1] = 1.

            # Apply the householder transformation
            self.householder.right_apply_transpose(self.Q_t[self.k:self.k+1, :])

        self.k += 1


    cpdef void update(UpdatingQT self, FLOAT_t[:] x):
        # Updates householder, then calls
        # update_qt

        # The Householder will detect if the new vector is linearly dependent on the previous
        # ones (within numerical precision specified by zero_tol).
        cdef bint dependent
        dependent = self.householder.update_from_column(x)

        # Mark the column as independent or dependent.  This information will be needed if the
        # column is ever downdated, since we then need to not downdate householder
        self.dependent_cols[self.k] = dependent

        # If linear dependence was detected, the householder will have failed to update
        # (as it should).  In that case, we want a row of zeros in our Q_t matrix because
        # the row space of Q_t should be the same as the span of all the x vectors passed to update.
        # A row of zeros makes this possible while still having self.k match the relevant dimension
        # of Q_t.  The update_qt method takes care of adding the zeros if dependent. Note this means
        # that in general self.householder.k <= self.k.  They are not necessarily equal.
        self.update_qt(dependent)



    cpdef void downdate(UpdatingQT self):
        self.k -= 1
        if not self.dependent_cols[self.k]:
            self.householder.downdate()

    cpdef void reset(UpdatingQT self):
        self.householder.reset()
        self.k = 0

cdef class Householder:

    def __init__(Householder self, int k, int m, int max_n,
                 FLOAT_t[::1, :] V, FLOAT_t[::1, :] T, FLOAT_t[::1] tau,
                 FLOAT_t[::1] beta, FLOAT_t[::1, :] work, FLOAT_t zero_tol):
        self.k = k
        self.m = m
        self.max_n = max_n
        self.V = V
        self.T = T
        self.tau = tau
        self.beta = beta
        self.work = work
        self.zero_tol = zero_tol

    @classmethod
    def alloc(cls, int m, int max_n, FLOAT_t zero_tol):
        cdef int k = 0
        cdef FLOAT_t[::1, :] V = np.empty(shape=(m, max_n), dtype=FLOAT, order='F')
        cdef FLOAT_t[::1, :] T = np.empty(shape=(max_n, max_n), dtype=FLOAT, order='F')
        cdef FLOAT_t[::1] tau = np.empty(shape=max_n, dtype=FLOAT, order='F')
        cdef FLOAT_t[::1] beta = np.empty(shape=max_n, dtype=FLOAT, order='F')
        cdef FLOAT_t[::1, :] work = np.empty(shape=(m, max_n), dtype=FLOAT, order='F')
        return cls(k, m, max_n, V, T, tau, beta, work, zero_tol)

    cpdef void downdate(Householder self):
        self.k -= 1

    cpdef void reset(Householder self):
        self.k = 0

    cpdef bint update_from_column(Householder self, FLOAT_t[:] c):
        # Copies c, applies self, then updates V and T

        # Copy c into V
        cdef int N = self.m
        cdef FLOAT_t * x = <FLOAT_t *> &(c[0])
        cdef int incx = c.strides[0] / c.itemsize
        cdef FLOAT_t * y = <FLOAT_t *> &(self.V[0, self.k])
        cdef int incy = 1
        dcopy(&N, x, &incx, y, &incy)

        # Apply self to new column in V
        self.left_apply_transpose(self.V[:, self.k:self.k+1])

        # Update V and T (increments k)
        return self.update_v_t()


    cpdef bint update_v_t(Householder self):
        # Assume relevant data has been copied into self.V correctly, as by
        # update_from_column.  Update V and T appropriately.
        cdef int n = self.m - self.k
        cdef FLOAT_t alpha = self.V[self.k, self.k]
        cdef FLOAT_t* x = <FLOAT_t *> &(self.V[(self.k + 1), self.k])
        cdef int incx = self.V.strides[0] // self.V.itemsize
        cdef FLOAT_t tau
        cdef FLOAT_t beta
        cdef bint dependent

        # Compute the householder reflection
        dlarfg(&n, &alpha, x, &incx, &tau)
        beta = alpha

        # If beta is very close to zero, the new column was linearly
        # dependent on the previous columns.  In that case, it's best if
        # we just pretend this never happened.  Note that this means k
        # will not be incremented.  UpdatingQT knows how to handle this
        # case, and will be informed by the return value.
        dependent = abs(beta) < self.zero_tol
        if dependent:
            return dependent

        # Add the new householder reflection to the
        # block reflector
        # TODO: Currently requires recalculating all of T
        # Could be updated to use BLAS instead to calculate
        # just the new column of T.  I'm not sure how to
        # do this or whether it would be faster.
        self.V[self.k, self.k] = 1.
        self.V[:self.k, self.k] = 0.
        self.tau[self.k] = tau
        self.beta[self.k] = alpha
        cdef char direct = 'F'
        cdef char storev = 'C'
        n = self.m
        cdef int k = self.k + 1
        cdef FLOAT_t * V = <FLOAT_t *> &(self.V[0,0])
        cdef int ldv = self.m
        cdef FLOAT_t * T = <FLOAT_t *> &(self.T[0,0])
        cdef FLOAT_t * tau_arg = <FLOAT_t *> &(self.tau[0])
        cdef int ldt = self.max_n
        dlarft(&direct, &storev, &n, &k, V, &ldv, tau_arg, T, &ldt)

        self.k += 1
        # Return beta in case the caller wants to diagnose linear dependence.
        return dependent

    cpdef void left_apply(Householder self, FLOAT_t[::1, :] C):
        cdef char side = 'L'
        cdef char trans = 'N'
        cdef char direct = 'F'
        cdef char storev = 'C'
        cdef int M = C.shape[0]
        cdef int N = C.shape[1]
        cdef int K = self.k
        cdef FLOAT_t * V = <FLOAT_t *> &(self.V[0, 0])
        cdef int ldv = self.m
        cdef FLOAT_t * T = <FLOAT_t *> &(self.T[0, 0])
        cdef int ldt = self.max_n
        cdef FLOAT_t * C_arg = <FLOAT_t *> &(C[0, 0])
        cdef int ldc = C.strides[1] // C.itemsize
        cdef FLOAT_t * work = <FLOAT_t *> &(self.work[0,0])
        cdef int ldwork = self.m
        print(C.shape)
        dlarfb(&side, &trans, &direct, &storev, &M, &N, &K,
               V, &ldv, T, &ldt, C_arg, &ldc, work, &ldwork)

    cpdef void left_apply_transpose(Householder self, FLOAT_t[::1, :] C):
        cdef char side = 'L'
        cdef char trans = 'T'
        cdef char direct = 'F'
        cdef char storev = 'C'
        cdef int M = C.shape[0]
        cdef int N = C.shape[1]
        cdef int K = self.k
        cdef FLOAT_t * V = <FLOAT_t *> &(self.V[0, 0])
        cdef int ldv = self.m
        cdef FLOAT_t * T = <FLOAT_t *> &(self.T[0, 0])
        cdef int ldt = self.max_n
        cdef FLOAT_t * C_arg = <FLOAT_t *> &(C[0, 0])
        cdef int ldc = C.strides[1] // C.itemsize
        cdef FLOAT_t * work = <FLOAT_t *> &(self.work[0,0])
        cdef int ldwork = self.m

        dlarfb(&side, &trans, &direct, &storev, &M, &N, &K,
               V, &ldv, T, &ldt, C_arg, &ldc, work, &ldwork)

    cpdef void right_apply(Householder self, FLOAT_t[::1, :] C):
        cdef char side = 'R'
        cdef char trans = 'N'
        cdef char direct = 'F'
        cdef char storev = 'C'
        cdef int M = C.shape[0]
        cdef int N = C.shape[1]
        cdef int K = self.k
        cdef FLOAT_t * V = <FLOAT_t *> &(self.V[0, 0])
        cdef int ldv = self.m
        cdef FLOAT_t * T = <FLOAT_t *> &(self.T[0, 0])
        cdef int ldt = self.max_n
        cdef FLOAT_t * C_arg = <FLOAT_t *> &(C[0, 0])
        cdef int ldc = C.strides[1] // C.itemsize
        cdef FLOAT_t * work = <FLOAT_t *> &(self.work[0,0])
        cdef int ldwork = self.m

        dlarfb(&side, &trans, &direct, &storev, &M, &N, &K,
               V, &ldv, T, &ldt, C_arg, &ldc, work, &ldwork)

    cpdef void right_apply_transpose(Householder self, FLOAT_t[::1, :] C):
        cdef char side = 'R'
        cdef char trans = 'T'
        cdef char direct = 'F'
        cdef char storev = 'C'
        cdef int M = C.shape[0]
        cdef int N = C.shape[1]
        cdef int K = self.k
        cdef FLOAT_t * V = <FLOAT_t *> &(self.V[0, 0])
        cdef int ldv = self.m
        cdef FLOAT_t * T = <FLOAT_t *> &(self.T[0, 0])
        cdef int ldt = self.max_n
        cdef FLOAT_t * C_arg = <FLOAT_t *> &(C[0, 0])
        cdef int ldc = C.strides[1] // C.itemsize
        cdef FLOAT_t * work = <FLOAT_t *> &(self.work[0,0])
        cdef int ldwork = self.m

        dlarfb(&side, &trans, &direct, &storev, &M, &N, &K,
               V, &ldv, T, &ldt, C_arg, &ldc, work, &ldwork)
#



from __future__ import annotations

import numpy
from numpy.testing import assert_almost_equal

from pyearth import Earth
from pyearth._basis import (
    Basis,
    ConstantBasisFunction,
    HingeBasisFunction,
    LinearBasisFunction,
)
from pyearth._types import BOOL
from pyearth.export import export_python_function

numpy.random.seed(0)

basis = Basis(10)
constant = ConstantBasisFunction()
basis.append(constant)
bf1 = HingeBasisFunction(constant, 0.1, 10, 1, False, "x1")
bf2 = HingeBasisFunction(constant, 0.1, 10, 1, True, "x1")
bf3 = LinearBasisFunction(bf1, 2, "x2")
basis.append(bf1)
basis.append(bf2)
basis.append(bf3)
X = numpy.random.normal(size=(100, 10))
missing = numpy.zeros_like(X, dtype=BOOL)
B = numpy.empty(shape=(100, 4), dtype=numpy.float64)
basis.transform(X, missing, B)
beta = numpy.random.normal(size=4)
y = numpy.empty(shape=100, dtype=numpy.float64)
y[:] = numpy.dot(B, beta) + numpy.random.normal(size=100)
beta2 = numpy.random.normal(size=4)
y2 = numpy.empty(shape=100, dtype=numpy.float64)
y2[:] = numpy.dot(B, beta2) + numpy.random.normal(size=100)
Y = numpy.concatenate([y[:, None], y2[:, None]], axis=1)
default_params = {"penalty": 1}


def test_export_python_function():
    for smooth in (True, False):
        model = Earth(penalty=1, smooth=smooth, max_degree=2).fit(X, y)
        export_model = export_python_function(model)
        for exp_pred, model_pred in zip(model.predict(X), export_model(X), strict=False):
            assert_almost_equal(exp_pred, model_pred)


# def test_export_python_string():
#     for smooth in (True, False):
#         model = Earth(penalty=1, smooth=smooth, max_degree=2).fit(X, y)
#         export_model = export_python_string(model, "my_test_model")
#         six.exec_(export_model, globals())
#         for exp_pred, model_pred in zip(model.predict(X), my_test_model(X)):  # noqa: F821
#             assert_almost_equal(exp_pred, model_pred)


# @if_pandas
# @if_sympy
# def test_export_sympy():
#     import pandas as pd
#     from sympy.printing.lambdarepr import NumPyPrinter
#     from sympy.utilities.lambdify import lambdify

#     class PyEarthNumpyPrinter(NumPyPrinter):
#         def _print_Max(self, expr):
#             return "maximum(" + ",".join(self._print(i) for i in expr.args) + ")"

#         def _print_NaNProtect(self, expr):
#             return (
#                 "where(isnan("
#                 + ",".join(self._print(a) for a in expr.args)
#                 + "), 0, "
#                 + ",".join(self._print(a) for a in expr.args)
#                 + ")"
#             )

#         def _print_Missing(self, expr):
#             return (
#                 "isnan("
#                 + ",".join(self._print(a) for a in expr.args)
#                 + ").astype(float)"
#             )

#     for smooth, n_cols, allow_missing in product((True, False), (1, 2), (True, False)):
#         X_df = pd.DataFrame(X.copy(), columns=["x_%d" % i for i in range(X.shape[1])])
#         y_df = pd.DataFrame(Y[:, :n_cols])
#         if allow_missing:
#             # Randomly remove some values so that the fitted model contains MissingnessBasisFunctions
#             X_df["x_1"][
#                 numpy.random.binomial(n=1, p=0.1, size=X_df.shape[0]).astype(bool)
#             ] = numpy.nan

#         model = Earth(allow_missing=allow_missing, smooth=smooth, max_degree=2).fit(
#             X_df, y_df
#         )
#         expressions = export_sympy(model) if n_cols > 1 else [export_sympy(model)]
#         module_dict = {
#             "select": numpy.select,
#             "less_equal": numpy.less_equal,
#             "isnan": numpy.isnan,
#             "greater_equal": numpy.greater_equal,
#             "logical_and": numpy.logical_and,
#             "less": numpy.less,
#             "logical_not": numpy.logical_not,
#             "greater": numpy.greater,
#             "maximum": numpy.maximum,
#             "Missing": lambda x: numpy.isnan(x).astype(float),
#             "NaNProtect": lambda x: numpy.where(numpy.isnan(x), 0, x),
#             "nan": numpy.nan,
#             "float": float,
#             "where": numpy.where,
#         }

#         for i, expression in enumerate(expressions):
#             # The lambdified functions for smoothed basis functions only work with modules='numpy' and
#             # for regular basis functions with modules={'Max':numpy.maximum}.  This is a confusing situation
#             func = lambdify(
#                 X_df.columns,
#                 expression,
#                 printer=PyEarthNumpyPrinter,
#                 modules=module_dict,
#             )
#             y_pred_sympy = func(*[X_df.loc[:, var] for var in X_df.columns])

#             y_pred = model.predict(X_df)[:, i] if n_cols > 1 else model.predict(X_df)
#             assert_array_almost_equal(y_pred, y_pred_sympy)
